# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import os
import pickle  # nosec  # pylint: disable=import_pickle
import tempfile
import uuid

import numpy as np

from .... import opcodes
from .... import tensor as mt
from ....lib.filesystem import get_fs
from ....core import OutputType
from ....core.context import get_context
from ....core.operand import OperandStage
from ....serialization.serializables import (
    StringField,
    Int32Field,
    Int64Field,
    DictField,
    BytesField,
    TupleField,
    DataTypeField,
)
from ....utils import has_unknown_shape, Timer
from ...operands import LearnOperand, LearnOperandMixin
from ..core import (
    proxima,
    get_proxima_type,
    validate_tensor,
    available_numpy_dtypes,
    rechunk_tensor,
    build_mmap_chunks,
)

logger = logging.getLogger(__name__)

DEFAULT_INDEX_SIZE = 5 * 10**6


class ProximaBuilder(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.PROXIMA_SIMPLE_BUILDER

    _distance_metric = StringField("distance_metric")
    _dimension = Int32Field("dimension")
    _column_number = Int64Field("column_number")
    _index_path = StringField("index_path")
    _index_builder = StringField("index_builder")
    _index_builder_params = DictField("index_builder_params")
    _index_converter = StringField("index_converter")
    _index_converter_params = DictField("index_converter_params")
    _topk = Int32Field("topk")
    _storage_options = BytesField(
        "storage_options", on_serialize=pickle.dumps, on_deserialize=pickle.loads
    )

    # only for chunk
    _array_shape = TupleField("array_shape")
    _array_dtype = DataTypeField("array_dtype")
    _offset = Int64Field("offset")

    def __init__(
        self,
        distance_metric=None,
        index_path=None,
        dimension=None,
        column_number=None,
        index_builder=None,
        index_builder_params=None,
        index_converter=None,
        index_converter_params=None,
        array_shape=None,
        array_dtype=None,
        offset=None,
        topk=None,
        storage_options=None,
        output_types=None,
        **kw,
    ):
        super().__init__(
            _distance_metric=distance_metric,
            _index_path=index_path,
            _dimension=dimension,
            _column_number=column_number,
            _index_builder=index_builder,
            _index_builder_params=index_builder_params,
            _array_shape=array_shape,
            _array_dtype=array_dtype,
            _offset=offset,
            _index_converter=index_converter,
            _index_converter_params=index_converter_params,
            _topk=topk,
            _storage_options=storage_options,
            _output_types=output_types,
            **kw,
        )
        if self._output_types is None:
            self._output_types = [OutputType.object]

    @property
    def distance_metric(self):
        return self._distance_metric

    @property
    def column_number(self):
        return self._column_number

    @property
    def index_path(self):
        return self._index_path

    @property
    def dimension(self):
        return self._dimension

    @property
    def index_builder(self):
        return self._index_builder

    @property
    def index_builder_params(self):
        return self._index_builder_params

    @property
    def index_converter(self):
        return self._index_converter

    @property
    def index_converter_params(self):
        return self._index_converter_params

    @property
    def topk(self):
        return self._topk

    @property
    def storage_options(self):
        return self._storage_options

    @property
    def array_shape(self):
        return self._array_shape

    @property
    def array_dtype(self):
        return self._array_dtype

    @property
    def offset(self):
        return self._offset

    def __call__(self, tensor):
        return self.new_tileable([tensor])

    @classmethod
    def _get_atleast_topk_nsplit(cls, nsplit, topk):
        new_nsplit = []
        i = 0
        while i < len(nsplit):
            cur = nsplit[i]
            i += 1
            if cur >= topk:
                new_nsplit.append(cur)
            else:
                while i < len(nsplit):
                    cur += nsplit[i]
                    i += 1
                    if cur >= topk:
                        break
                if cur < topk and len(new_nsplit) > 0:
                    new_nsplit[-1] += cur
                elif cur >= topk:
                    new_nsplit.append(cur)
        new_nsplit = tuple(new_nsplit)
        assert sum(new_nsplit) == sum(
            nsplit
        ), f"sum of nsplit not equal, old: {nsplit}, new: {new_nsplit}"

        return new_nsplit

    @classmethod
    def tile(cls, op):
        tensor = op.inputs[0]
        out = op.outputs[0]
        index_path = op.index_path
        ctx = get_context()
        fs = None
        if index_path is not None:
            fs = get_fs(index_path, op.storage_options)

        if index_path is not None:
            # check if the index path is empty
            try:
                files = [f for f in fs.ls(index_path) if "proxima_" in f]
                if files:
                    raise ValueError(
                        f"Directory {index_path} contains built proxima index, "
                        f"clean them to perform new index building"
                    )
            except FileNotFoundError:
                # if not exist, create directory
                fs.mkdir(index_path)

        # make sure all inputs have known chunk sizes
        if has_unknown_shape(*op.inputs):
            yield

        if op.column_number:
            index_chunk_size = op.inputs[0].shape[0] // op.column_number
        else:
            worker_num = len(ctx.get_worker_addresses() or [])
            if worker_num > 0:
                index_chunk_size = max(
                    op.inputs[0].shape[0] // worker_num, DEFAULT_INDEX_SIZE
                )
            else:
                index_chunk_size = DEFAULT_INDEX_SIZE

        if op.topk is not None:
            index_chunk_size = cls._get_atleast_topk_nsplit(index_chunk_size, op.topk)

        # build chunks for writing tensors to mmap files.
        worker_iter = iter(itertools.cycle(ctx.get_worker_addresses() or [None]))
        chunk_groups = rechunk_tensor(tensor, index_chunk_size)
        out_chunks = []
        offsets = []
        offset = 0
        for chunk_group in chunk_groups:
            offsets.append(offset)
            file_prefix = f"proxima-build-{str(uuid.uuid4())}"
            out_chunks.append(
                build_mmap_chunks(
                    chunk_group, next(worker_iter), file_prefix=file_prefix
                )
            )
            offset += sum(c.shape[0] for c in chunk_group)

        final_out_chunks = []
        for j, chunks in enumerate(out_chunks):
            chunk_op = op.copy().reset_key()
            chunk_op.stage = OperandStage.map
            chunk_op.expect_worker = chunks[0].op.expect_worker
            chunk_op._array_shape = chunks[0].op.total_shape
            chunk_op._array_dtype = chunks[0].dtype
            chunk_op._offset = offsets[j]
            out_chunk = chunk_op.new_chunk(chunks, index=(j,))
            final_out_chunks.append(out_chunk)

        logger.warning(f"index chunks count: {len(final_out_chunks)} ")

        params = out.params
        params["chunks"] = final_out_chunks
        params["nsplits"] = ((1,) * len(final_out_chunks),)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _execute_map(cls, ctx, op: "ProximaBuilder"):
        mmap_path = ctx[op.inputs[0].key]
        out = op.outputs[0]

        data = np.memmap(
            mmap_path, dtype=op.array_dtype, mode="r", shape=op.array_shape
        )

        proxima_type = get_proxima_type(op.array_dtype)
        offset = op.offset

        # holder
        with Timer() as timer:
            holder = proxima.IndexHolder(
                type=proxima_type, dimension=op.dimension, shallow=True
            )
            holder.mount(data, key_base=offset)

        logger.warning(f"Holder({op.key}) costs {timer.duration} seconds")

        # converter
        meta = proxima.IndexMeta(
            proxima_type, dimension=op.dimension, measure_name=op.distance_metric
        )
        if op.index_converter is not None:
            with Timer() as timer:
                converter = proxima.IndexConverter(
                    name=op.index_converter, meta=meta, params=op.index_converter_params
                )
                converter.train_and_transform(holder)
                holder = converter.result()
                meta = converter.meta()

            logger.warning(f"Converter({op.key}) costs {timer.duration} seconds")

        # builder
        with Timer() as timer:
            builder = proxima.IndexBuilder(
                name=op.index_builder, meta=meta, params=op.index_builder_params
            )
            builder = builder.train_and_build(holder)

        logger.warning(f"Builder({op.key}) costs {timer.duration} seconds")

        # remove mmap file
        os.remove(mmap_path)

        # dumper
        with Timer() as timer:
            path = tempfile.mkstemp(prefix="proxima-", suffix=".index")[1]
            dumper = proxima.IndexDumper(name="FileDumper", path=path)
            builder.dump(dumper)
            dumper.close()

        logger.warning(f"Dumper({op.key}) costs {timer.duration} seconds")

        if op.index_path is None:
            ctx[out.key] = path
        else:
            # write to external file
            with Timer() as timer:
                fs = get_fs(op.index_path, op.storage_options)
                filename = f"proxima_{out.index[0]}_index"
                out_path = f'{op.index_path.rstrip("/")}/{filename}'

                def write_index():
                    with fs.open(out_path, "wb") as out_f:
                        with open(path, "rb") as in_f:
                            # 128M
                            chunk_bytes = 128 * 1024**2
                            while True:
                                data = in_f.read(chunk_bytes)
                                if data:
                                    out_f.write(data)
                                else:
                                    break

                # retry 3 times
                for _ in range(3):
                    try:
                        write_index()
                        break
                    except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                        fs.delete(out_path)
                        continue

            logger.warning(
                f"WritingToVolume({op.key}), out path: {out_path}, "
                f"size {os.path.getsize(path)}, "
                f"costs {timer.duration} seconds "
                f"speed {round(os.path.getsize(path) / (1024 ** 2) / timer.duration, 2)} MB/s"
            )

            ctx[out.key] = filename

    @classmethod
    def _execute_agg(cls, ctx, op: "ProximaBuilder"):
        paths = [ctx[inp.key] for inp in op.inputs]
        ctx[op.outputs[0].key] = paths

    @classmethod
    def execute(cls, ctx, op: "ProximaBuilder"):
        if op.stage != OperandStage.agg:
            return cls._execute_map(ctx, op)
        else:
            return cls._execute_agg(ctx, op)

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        assert not tileable.is_coarse()

        op = cls(stage=OperandStage.agg)
        chunk = cls(stage=OperandStage.agg).new_chunk(tileable.chunks)
        return op.new_tileable([tileable], chunks=[chunk], nsplits=((1,),))


def build_index(
    tensor,
    dimension=None,
    index_path=None,
    column_number=None,
    need_shuffle=False,
    distance_metric="SquaredEuclidean",
    index_builder="SsgBuilder",
    index_builder_params=None,
    index_converter=None,
    index_converter_params=None,
    topk=None,
    storage_options=None,
    run=True,
    session=None,
    run_kwargs=None,
):
    tensor = validate_tensor(tensor)
    if tensor.dtype not in available_numpy_dtypes:
        raise ValueError(
            f"Dtype to build index should be one of {available_numpy_dtypes}, "
            f"got {tensor.dtype}"
        )

    if dimension is None:
        dimension = tensor.shape[1]
    if index_builder_params is None:
        index_builder_params = {}
    if index_converter_params is None:
        index_converter_params = {}

    if need_shuffle:
        tensor = mt.random.permutation(tensor)

    op = ProximaBuilder(
        distance_metric=distance_metric,
        index_path=index_path,
        dimension=dimension,
        column_number=column_number,
        index_builder=index_builder,
        index_builder_params=index_builder_params,
        index_converter=index_converter,
        index_converter_params=index_converter_params,
        topk=topk,
        storage_options=storage_options,
    )
    result = op(tensor)
    if run:
        return result.execute(session=session, **(run_kwargs or dict()))
    else:
        return result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import random
import struct
import sys
from string import ascii_letters, digits
from typing import Any, Dict, List, Tuple, Optional
from dataclasses import dataclass

try:
    if sys.version_info[:2] >= (3, 8):
        # builtin package for Python 3.8+
        from multiprocessing.shared_memory import SharedMemory
    else:
        # backport package for Python 3.7-
        from shared_memory import SharedMemory

    class SharedMemoryForRead(SharedMemory):
        def __del__(self):
            # close fd only
            fd = self._fd
            if os.name != "nt" and fd >= 0:
                os.close(fd)

except ImportError:  # pragma: no cover
    # allow shared_memory package to be absent
    SharedMemory = SharedMemoryForRead = None

from ..serialization import AioSerializer, AioDeserializer
from ..utils import implements, dataslots
from .base import StorageBackend, StorageLevel, ObjectInfo, register_storage_backend
from .core import BufferWrappedFileObject, StorageFileObject

_is_windows: bool = sys.platform.startswith("win")
_qword_pack = struct.Struct("<Q")


@dataslots
@dataclass
class WinShmObjectInfo(ObjectInfo):
    shm: Any = None


class SharedMemoryFileObject(BufferWrappedFileObject):
    def __init__(self, object_id: Any, mode: str, size: Optional[int] = None):
        self.shm = None
        super().__init__(object_id, mode, size=size)

    def _write_actual_size(self):
        # we need to reopen the SharedMemory object as the size
        # of the original one is less than the actual size.
        actual_shm = SharedMemory(name=self._object_id)
        actual_shm.buf[-8:] = _qword_pack.pack(self._size)

    def _write_init(self):
        # keep last 8 bytes to record actual memory size
        self.shm = shm = SharedMemory(
            name=self._object_id, create=True, size=self._size + 8
        )
        self._write_actual_size()
        self._buffer = self._mv = shm.buf

    def _read_init(self):
        self.shm = shm = SharedMemoryForRead(name=self._object_id)
        self._buffer = self._mv = shm.buf
        if self._size is None:
            (self._size,) = _qword_pack.unpack(shm.buf[-8:])

    def _write_close(self):
        pass

    def _read_close(self):
        pass


class ShmStorageFileObject(StorageFileObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._shm = None

    async def close(self):
        if _is_windows:
            self._shm = self._file.shm
        await super().close()


@register_storage_backend
class SharedMemoryStorage(StorageBackend):
    name = "shared_memory"

    def __init__(self, **kw):
        if kw:  # pragma: no cover
            raise TypeError(
                f'SharedMemoryStorage got unexpected arguments: {",".join(kw)}'
            )
        # for test purpose, in real usage,
        # each storage object holds different object ids,
        # we cannot do any operation according to
        # this property only
        self._object_ids = set()

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        if kwargs:  # pragma: no cover
            raise TypeError(
                f'SharedMemoryStorage got unexpected config: {",".join(kwargs)}'
            )

        return dict(), dict()

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        object_ids = kwargs.get("object_ids") or ()
        for object_id in object_ids:
            try:
                shm = SharedMemory(name=object_id)
                shm.unlink()
                await asyncio.sleep(0)
            except FileNotFoundError:
                pass

    @property
    @implements(StorageBackend.level)
    def level(self) -> StorageLevel:
        return StorageLevel.MEMORY

    @classmethod
    def _generate_object_id(cls):
        return "".join(random.choice(ascii_letters + digits) for _ in range(30))

    @implements(StorageBackend.get)
    async def get(self, object_id, **kwargs) -> object:
        if kwargs:  # pragma: no cover
            raise NotImplementedError(f'Got unsupported args: {",".join(kwargs)}')

        shm_file = SharedMemoryFileObject(object_id, mode="r")

        async with StorageFileObject(shm_file, object_id) as f:
            deserializer = AioDeserializer(f)
            return await deserializer.run()

    @implements(StorageBackend.put)
    async def put(self, obj, importance=0) -> ObjectInfo:
        object_id = self._generate_object_id()

        serializer = AioSerializer(obj)
        buffers = await serializer.run()
        buffer_size = sum(getattr(buf, "nbytes", len(buf)) for buf in buffers)

        shm_file = SharedMemoryFileObject(object_id, mode="w", size=buffer_size)
        async with StorageFileObject(shm_file, object_id) as f:
            for buffer in buffers:
                await f.write(buffer)

        self._object_ids.add(object_id)
        if _is_windows:
            return WinShmObjectInfo(
                size=buffer_size, object_id=object_id, shm=shm_file.shm
            )
        else:
            return ObjectInfo(size=buffer_size, object_id=object_id)

    @implements(StorageBackend.delete)
    async def delete(self, object_id):
        try:
            shm = SharedMemory(name=object_id)
            shm.unlink()
            shm.close()
        except FileNotFoundError:
            if sys.platform == "win32":
                # skip file not found error for windows
                pass
            else:  # pragma: no cover
                raise
        try:
            self._object_ids.remove(object_id)
        except KeyError:  # pragma: no cover
            return

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id) -> ObjectInfo:
        shm_file = SharedMemoryFileObject(object_id, mode="r")

        async with ShmStorageFileObject(shm_file, object_id) as f:
            deserializer = AioDeserializer(f)
            size = await deserializer.get_size()
        if not _is_windows:
            return ObjectInfo(size=size, object_id=object_id)
        else:
            return WinShmObjectInfo(size=size, object_id=object_id, shm=shm_file)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        if size is None:  # pragma: no cover
            raise ValueError("size must be provided for shared memory backend")

        new_id = self._generate_object_id()
        shm_file = SharedMemoryFileObject(new_id, size=size, mode="w")
        return ShmStorageFileObject(shm_file, object_id=new_id)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        shm_file = SharedMemoryFileObject(object_id, mode="r")
        return ShmStorageFileObject(shm_file, object_id=object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:  # pragma: no cover
        raise NotImplementedError("Shared memory storage does not support list")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...utils import lazy_import
from ...serialization.serializables import Int32Field, ListField, FieldTypes
from ..operands import DataFrameOperandMixin, DataFrameOperand


cudf = lazy_import("cudf")


class ChunkStandardizeRangeIndex(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.STANDARDIZE_RANGE_INDEX

    axis = Int32Field("axis")
    prev_shapes = ListField("prev_shapes", FieldTypes.tuple)

    @classmethod
    def execute(cls, ctx, op: "ChunkStandardizeRangeIndex"):
        xdf = cudf if op.gpu else pd
        in_data = ctx[op.inputs[0].key].copy()
        index_start = sum([shape[op.axis] for shape in op.prev_shapes])
        if op.axis == 0:
            in_data.index = xdf.RangeIndex(index_start, index_start + len(in_data))
        else:
            in_data.columns = xdf.RangeIndex(
                index_start, index_start + in_data.shape[1]
            )
        ctx[op.outputs[0].key] = in_data



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List

from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from .core import AbstractLifecycleAPI


class LifecycleWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = "/api/session/(?P<session_id>[^/]+)/lifecycle"

    async def _get_oscar_lifecycle_api(self, session_id: str):
        from .oscar import LifecycleAPI

        return await self._get_api_by_key(LifecycleAPI, session_id)

    @web_api("", method="post", arg_filter={"action": "decref_tileables"})
    async def decref_tileables(self, session_id: str):
        tileable_keys = self.get_argument("tileable_keys").split(",")
        counts = self.get_argument("counts", None)
        if counts:
            counts = [int(c) for c in counts.split(",")]

        oscar_api = await self._get_oscar_lifecycle_api(session_id)
        await oscar_api.decref_tileables(tileable_keys, counts=counts)

    @web_api("", method="get", arg_filter={"action": "get_all_chunk_ref_counts"})
    async def get_all_chunk_ref_counts(self, session_id: str):
        oscar_api = await self._get_oscar_lifecycle_api(session_id)
        res = await oscar_api.get_all_chunk_ref_counts()
        self.write(serialize_serializable(res))


web_handlers = {LifecycleWebAPIHandler.get_root_pattern(): LifecycleWebAPIHandler}


class WebLifecycleAPI(AbstractLifecycleAPI, MarsWebAPIClientMixin):
    def __init__(
        self, session_id: str, address: str, request_rewriter: Callable = None
    ):
        self._session_id = session_id
        self._address = address.rstrip("/")
        self.request_rewriter = request_rewriter

    async def decref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        path = f"{self._address}/api/session/{self._session_id}/lifecycle"
        params = dict(action="decref_tileables")
        counts = (
            f"&counts={','.join(str(c) for c in counts)}" if counts is not None else ""
        )
        await self._request_url(
            path=path,
            method="POST",
            params=params,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data="tileable_keys=" + ",".join(tileable_keys) + counts,
        )

    async def get_all_chunk_ref_counts(self) -> Dict[str, int]:
        params = dict(action="get_all_chunk_ref_counts")
        path = f"{self._address}/api/session/{self._session_id}/lifecycle"
        res = await self._request_url("GET", path, params=params)
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import numpy as np

from ...typing import TileableType
from .. import asarray


def in1d(
    ar1: Union[TileableType, np.ndarray],
    ar2: Union[TileableType, np.ndarray, list],
    assume_unique: bool = False,
    invert: bool = False,
):
    """
    Test whether each element of a 1-D tensor is also present in a second tensor.

    Returns a boolean tensor the same length as `ar1` that is True
    where an element of `ar1` is in `ar2` and False otherwise.

    We recommend using :func:`isin` instead of `in1d` for new code.

    Parameters
    ----------
    ar1 : (M,) Tensor
        Input tensor.
    ar2 : array_like
        The values against which to test each value of `ar1`.
    assume_unique : bool, optional
        If True, the input tensors are both assumed to be unique, which
        can speed up the calculation.  Default is False.
    invert : bool, optional
        If True, the values in the returned tensor are inverted (that is,
        False where an element of `ar1` is in `ar2` and True otherwise).
        Default is False. ``np.in1d(a, b, invert=True)`` is equivalent
        to (but is faster than) ``np.invert(in1d(a, b))``.

    Returns
    -------
    in1d : (M,) Tensor, bool
        The values `ar1[in1d]` are in `ar2`.

    See Also
    --------
    isin                  : Version of this function that preserves the
                            shape of ar1.
    numpy.lib.arraysetops : Module with a number of other functions for
                            performing set operations on arrays.

    Notes
    -----
    `in1d` can be considered as an element-wise function version of the
    python keyword `in`, for 1-D sequences. ``in1d(a, b)`` is roughly
    equivalent to ``mt.array([item in b for item in a])``.
    However, this idea fails if `ar2` is a set, or similar (non-sequence)
    container:  As ``ar2`` is converted to a tensor, in those cases
    ``asarray(ar2)`` is an object tensor rather than the expected tensor of
    contained values.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> test = mt.array([0, 1, 2, 5, 0])
    >>> states = [0, 2]
    >>> mask = mt.in1d(test, states)
    >>> mask.execute()
    array([ True, False,  True, False,  True])
    >>> test[mask].execute()
    array([0, 2, 0])
    >>> mask = mt.in1d(test, states, invert=True)
    >>> mask.execute()
    array([False,  True, False,  True, False])
    >>> test[mask].execute()
    array([1, 5])
    """
    from .isin import isin

    ar1 = asarray(ar1).ravel()
    ar2 = asarray(ar2).ravel()
    return isin(ar1, ar2, assume_unique=assume_unique, invert=invert)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from ..core import StorageManagerActor


class StorageWorkerService(AbstractService):
    """
    Storage service on worker

    Service Configuration
    ---------------------
    {
        "storage": {
            "backends": ["plasma"],
            "<storage backend name>"： "<setup params>",
        }
    }
    """

    async def start(self):
        storage_configs = self._config["storage"]
        backends = storage_configs.get("backends")
        options = storage_configs.get("default_config", dict())
        transfer_block_size = options.get("transfer_block_size", None)
        backend_config = {}
        for backend in backends:
            storage_config = storage_configs.get(backend, dict())
            backend_config[backend] = storage_config
            if backend == "ray":
                # Specify supervisor as ray owner will be costly when mars do shuffle which there will be m*n objects
                # need to specify supervisor as owner, so enable it only for auto scale to avoid data lost when scale
                # in. This limit can be removed when ray support ownership transfer.
                if (
                    self._config.get("scheduling", {})
                    .get("autoscale", {})
                    .get("enabled", False)
                ):
                    try:
                        from ...cluster.api import ClusterAPI

                        cluster_api = await ClusterAPI.create(self._address)
                        supervisor_address = (await cluster_api.get_supervisors())[0]
                        # ray storage backend need to set supervisor as owner to avoid data lost when worker dies.
                        owner = supervisor_address
                    except mo.ActorNotExist:
                        owner = self._address
                else:
                    owner = self._address
                storage_config["owner"] = owner

        await mo.create_actor(
            StorageManagerActor,
            backend_config,
            transfer_block_size,
            uid=StorageManagerActor.default_uid(),
            address=self._address,
        )

    async def stop(self):
        await mo.destroy_actor(
            mo.create_actor_ref(
                address=self._address, uid=StorageManagerActor.default_uid()
            )
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Union

from sklearn.metrics import make_scorer

from . import accuracy_score, log_loss, r2_score


accuracy_score = make_scorer(accuracy_score)
r2_score = make_scorer(r2_score)
neg_log_loss_scorer = make_scorer(log_loss, greater_is_better=False, needs_proba=True)


SCORERS = dict(
    r2=r2_score,
    accuracy=accuracy_score,
    neg_log_loss=neg_log_loss_scorer,
)


def get_scorer(score_func: Union[str, Callable], **kwargs) -> Callable:
    """
    Get a scorer from string

    Parameters
    ----------
    score_func : str | callable
        scoring method as string. If callable it is returned as is.

    Returns
    -------
    scorer : callable
        The scorer.
    """
    if isinstance(score_func, str):
        try:
            scorer = SCORERS[score_func]
        except KeyError:
            raise ValueError(
                "{} is not a valid scoring value. "
                "Valid options are {}".format(score_func, sorted(SCORERS))
            )
        return scorer
    else:
        return make_scorer(score_func, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...serialization.serializables import Int32Field, Float64Field
from ..datasource import arange
from ..operands import TensorOperand, TensorOperandMixin
from ..core import TensorOrder


class TensorRFFTFreq(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.RFFTFREQ

    _n = Int32Field("n")
    _d = Float64Field("d")

    def __init__(self, n=None, d=None, **kw):
        super().__init__(_n=n, _d=d, **kw)

    @property
    def n(self):
        return self._n

    @property
    def d(self):
        return self._d

    def __call__(self, chunk_size=None):
        shape = (self.n // 2 + 1,)
        return self.new_tensor(
            None, shape, raw_chunk_size=chunk_size, order=TensorOrder.C_ORDER
        )

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        t = arange(
            tensor.shape[0],
            dtype=op.dtype,
            gpu=op.gpu,
            chunk_size=tensor.extra_params.raw_chunk_size,
        )
        t = t / (op.n * op.d)
        t = yield from recursive_tile(t)

        new_op = op.copy()
        return new_op.new_tensors(
            None,
            tensor.shape,
            order=tensor.order,
            chunks=t.chunks,
            nsplits=t.nsplits,
            **tensor.extra_params
        )


def rfftfreq(n, d=1.0, gpu=None, chunk_size=None):
    """
    Return the Discrete Fourier Transform sample frequencies
    (for usage with rfft, irfft).

    The returned float tensor `f` contains the frequency bin centers in cycles
    per unit of the sample spacing (with zero at the start).  For instance, if
    the sample spacing is in seconds, then the frequency unit is cycles/second.

    Given a window length `n` and a sample spacing `d`::

      f = [0, 1, ...,     n/2-1,     n/2] / (d*n)   if n is even
      f = [0, 1, ..., (n-1)/2-1, (n-1)/2] / (d*n)   if n is odd

    Unlike `fftfreq` (but like `scipy.fftpack.rfftfreq`)
    the Nyquist frequency component is considered to be positive.

    Parameters
    ----------
    n : int
        Window length.
    d : scalar, optional
        Sample spacing (inverse of the sampling rate). Defaults to 1.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    f : Tensor
        Tensor of length ``n//2 + 1`` containing the sample frequencies.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> signal = mt.array([-2, 8, 6, 4, 1, 0, 3, 5, -3, 4], dtype=float)
    >>> fourier = mt.fft.rfft(signal)
    >>> n = signal.size
    >>> sample_rate = 100
    >>> freq = mt.fft.fftfreq(n, d=1./sample_rate)
    >>> freq.execute()
    array([  0.,  10.,  20.,  30.,  40., -50., -40., -30., -20., -10.])
    >>> freq = mt.fft.rfftfreq(n, d=1./sample_rate)
    >>> freq.execute()
    array([  0.,  10.,  20.,  30.,  40.,  50.])

    """
    n, d = int(n), float(d)
    op = TensorRFFTFreq(n=n, d=d, dtype=np.dtype(float), gpu=gpu)
    return op(chunk_size=chunk_size)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Backport of the asyncio.runners module from Python 3.7.
"""
# Source:
# https://github.com/python/cpython/blob/a4afcdfa55ddffa4b9ae3b0cf101628c7bff4102/Lib/asyncio/runners.py

# Modifications:
# * removed relative imports of .coroutines, .events, .tasks
# * replaced `coroutines`, `events`, `tasks` with `asyncio`.
# * replaced `tasks.all_tasks` with `asyncio.Task.all_tasks` because it is
#   backwards compatible.
# * Use private function `asyncio.events._get_running_loop` directly in
#   Python 3.6

import asyncio
import weakref
from typing import Any, Awaitable, Coroutine, TypeVar, Union


try:
    from asyncio import get_running_loop  # noqa Python >=3.7
except ImportError:  # pragma: no cover
    from asyncio.events import _get_running_loop as get_running_loop  # pragma: no cover

__all__ = ("run", "get_running_loop")
_T = TypeVar("_T")


def _patch_loop(loop):
    """
    This function is designed to work around https://bugs.python.org/issue36607

    It's job is to keep a thread safe variable tasks up to date with any tasks that
    are created for the given loop. This then lets you cancel them as _all_tasks
    was intended for.

    We also need to patch the {get,set}_task_factory functions because we can't allow
    Other users of it to overwrite our factory function. This function will pretend
    like there is no factory set but in reality our factory is always set and we will
    call the provided one set
    """
    tasks = weakref.WeakSet()

    task_factory = [None]

    def _set_task_factory(factory):
        task_factory[0] = factory

    def _get_task_factory():
        return task_factory[0]

    def _safe_task_factory(loop, coro):
        if task_factory[0] is None:
            # These lines are copied from the standard library because they don't have
            # this inside a default factory function for me to call.
            # https://github.com/python/cpython/blob/3.6/Lib/asyncio/base_events.py#L304
            task = asyncio.Task(coro, loop=loop)
            if task._source_traceback:
                del task._source_traceback[-1]  # pragma: no cover
        else:
            task = task_factory[0](loop, coro)
        tasks.add(task)
        return task

    loop.set_task_factory(_safe_task_factory)
    loop.set_task_factory = _set_task_factory
    loop.get_task_factory = _get_task_factory

    return tasks


def run(
    main: Union[Coroutine[Any, None, _T], Awaitable[_T]], *, debug: bool = False
) -> _T:
    """Run a coroutine.

    This function runs the passed coroutine, taking care of
    managing the asyncio event loop and finalizing asynchronous
    generators.

    This function cannot be called when another asyncio event loop is
    running in the same thread.

    If debug is True, the event loop will be run in debug mode.

    This function always creates a new event loop and closes it at the end.
    It should be used as a main entry point for asyncio programs, and should
    ideally only be called once.

    Example:

        async def main():
            await asyncio.sleep(1)
            print('hello')

        asyncio.run(main())
    """
    # Python 3.7+ raises RuntimeError while <3.6 returns None
    try:
        loop = get_running_loop()
    except RuntimeError:
        loop = None
    if loop is not None:
        raise RuntimeError("asyncio.run() cannot be called from a running event loop")

    if not asyncio.iscoroutine(main):
        raise ValueError("a coroutine was expected, got {!r}".format(main))

    loop = asyncio.new_event_loop()
    tasks = _patch_loop(loop)

    try:
        asyncio.set_event_loop(loop)
        loop.set_debug(debug)
        return loop.run_until_complete(main)
    finally:
        try:
            _cancel_all_tasks(loop, tasks)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            asyncio.set_event_loop(None)  # type: ignore
            loop.close()


def _cancel_all_tasks(loop, tasks):
    to_cancel = [task for task in tasks if not task.done()]

    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(
        asyncio.gather(*to_cancel, loop=loop, return_exceptions=True)
    )

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ...core import OutputType, recursive_tile
from ...serialization.serializables import StringField, BoolField, AnyField, ListField
from ..core import SERIES_TYPE
from ..datasource.dataframe import from_pandas as from_pandas_df
from ..datasource.series import from_pandas as from_pandas_series
from ..initializer import Series as asseries
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..reduction.unique import unique
from ..utils import gen_unknown_index_value

_encoding_dtype_kind = ["O", "S", "U"]


class DataFrameGetDummies(DataFrameOperand, DataFrameOperandMixin):
    prefix = AnyField("prefix")
    prefix_sep = StringField("prefix_sep")
    dummy_na = BoolField("dummy_na")
    columns = ListField("columns")
    sparse = BoolField("sparse")
    drop_first = BoolField("drop_first")
    dtype = AnyField("dtype")

    def __init__(
        self,
        prefix=None,
        prefix_sep=None,
        dummy_na=None,
        columns=None,
        sparse=None,
        drop_first=None,
        dtype=None,
        **kws,
    ):
        super().__init__(
            prefix=prefix,
            prefix_sep=prefix_sep,
            dummy_na=dummy_na,
            columns=columns,
            sparse=sparse,
            drop_first=drop_first,
            dtype=dtype,
            **kws,
        )
        self.output_types = [OutputType.dataframe]

    @classmethod
    def tile(cls, op):
        inp = op.inputs[0]
        out = op.outputs[0]
        if len(inp.chunks) == 1:
            chunk_op = op.copy().reset_key()
            chunk_param = out.params
            chunk_param["index"] = (0, 0)
            chunk = chunk_op.new_chunk(inp.chunks, kws=[chunk_param])
            new_op = op.copy().reset_key()
            param = out.params
            param["chunks"] = [chunk]
            param["nsplits"] = ((np.nan,), (np.nan,))
            return new_op.new_dataframe(op.inputs, kws=[param])
        elif isinstance(inp, SERIES_TYPE):
            unique_inp = yield from recursive_tile(unique(inp))
            chunks = []
            for c in inp.chunks:
                chunk_op = op.copy().reset_key()
                chunk_param = out.params
                chunk_param["index_value"] = gen_unknown_index_value(c.index_value)
                chunk_param["index"] = (c.index[0], 0)
                chunk = chunk_op.new_chunk([c] + unique_inp.chunks, kws=[chunk_param])
                chunks.append(chunk)

            new_op = op.copy().reset_key()
            param = out.params
            param["chunks"] = chunks
            param["nsplits"] = (tuple([np.nan] * inp.chunk_shape[0]), (np.nan,))
            return new_op.new_dataframe(op.inputs, kws=[param])
        else:
            if op.columns:
                encoding_columns = op.columns
            else:
                encoding_columns = []
                for idx, dtype in enumerate(inp.dtypes.values):
                    if dtype.kind in _encoding_dtype_kind:
                        column = inp.dtypes.index[idx]
                        encoding_columns.append(column)
            # reindex, make encoding columns in the end of dataframe, to keep pace with pandas.get_dummies
            total_columns = list(inp.columns.to_pandas().array)
            for col in encoding_columns:
                total_columns.remove(col)
            total_columns.extend(encoding_columns)
            inp = yield from recursive_tile(inp[total_columns])

            unique_chunks = dict()
            for col in encoding_columns:
                unique_chunks[col] = yield from recursive_tile(unique(inp[col]))

            chunks = []
            prefix = op.prefix
            column_to_prefix = dict()
            for c in inp.chunks:
                chunk_op = op.copy().reset_key()
                chunk_op.columns = []
                if isinstance(chunk_op.prefix, list):
                    chunk_op.prefix = []
                chunk_param = c.params
                chunk_param["shape"] = (np.nan, np.nan)
                chunk_columns = c.dtypes.index
                inp_chunk = [c]
                for chunk_column in chunk_columns:
                    if chunk_column in encoding_columns:
                        chunk_op.columns.append(chunk_column)
                        inp_chunk.extend(unique_chunks[chunk_column].chunks)
                        if isinstance(prefix, list):
                            if chunk_column in column_to_prefix.keys():
                                chunk_op.prefix.append(column_to_prefix[chunk_column])
                            else:
                                column_to_prefix[chunk_column] = prefix[0]
                                chunk_op.prefix.append(prefix[0])
                                prefix = prefix[1:]
                chunk = chunk_op.new_chunk(inp_chunk, kws=[chunk_param])
                chunks.append(chunk)

            new_op = op.copy()
            kw = out.params.copy()
            kw["chunks"] = chunks
            kw["nsplits"] = (
                tuple([np.nan] * inp.chunk_shape[0]),
                tuple([np.nan] * inp.chunk_shape[1]),
            )
            return new_op.new_dataframe(op.inputs, kws=[kw])

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.inputs[0].key]
        result_length = inp.shape[0]
        unique_inputs = []
        for unique_input in op.inputs[1:]:
            unique_inputs.append(ctx[unique_input.key].tolist())

        if unique_inputs:
            if isinstance(inp, pd.Series):
                extra_series = pd.Series(unique_inputs[0])
                inp = pd.concat([inp, extra_series])
            else:
                # make all unique_input's length the same, then get a dataframe
                max_length = len(max(unique_inputs, key=len))
                unique_inputs = [
                    unique_list + [unique_list[0]] * (max_length - len(unique_list))
                    for unique_list in unique_inputs
                ]
                extra_dataframe = pd.DataFrame(dict(zip(op.columns, unique_inputs)))

                # add the columns that need not to encode, to concat extra_dataframe and inp
                total_columns = list(inp.columns.array)
                for col in op.columns:
                    total_columns.remove(col)
                remain_columns = total_columns
                not_encode_columns = []
                if len(remain_columns) > 0:
                    for col in remain_columns:
                        not_encode_columns.append([inp[col].iloc[0]] * max_length)
                not_encode_dataframe = pd.DataFrame(
                    dict(zip(remain_columns, not_encode_columns))
                )

                extra_dataframe = pd.concat(
                    [not_encode_dataframe, extra_dataframe], axis=1
                )
                inp = pd.concat([inp, extra_dataframe], axis=0)

        result = pd.get_dummies(
            inp,
            op.prefix,
            op.prefix_sep,
            op.dummy_na,
            op.columns,
            op.sparse,
            op.drop_first,
            op.dtype,
        )
        ctx[op.outputs[0].key] = result.iloc[:result_length]

    def __call__(self, data):
        if isinstance(data, (list, tuple)):
            data = asseries(data)
        elif isinstance(data, pd.Series):
            data = from_pandas_series(data)
        elif isinstance(data, pd.DataFrame):
            data = from_pandas_df(data)

        if self.prefix is not None:
            if isinstance(self.prefix, list):
                if self.columns is not None:
                    encoding_col_num = len(self.columns)
                else:
                    encoding_col_num = 0
                    for dtype in data.dtypes.values:
                        if dtype.kind in _encoding_dtype_kind:
                            encoding_col_num += 1
                prefix_num = len(self.prefix)
                if prefix_num != encoding_col_num:
                    raise ValueError(
                        f"Length of 'prefix' ({prefix_num}) did not match "
                        + f"the length of the columns being encoded ({encoding_col_num})"
                    )
            elif isinstance(self.prefix, dict):
                if self.columns is not None:
                    encoding_col_num = len(self.columns)
                    prefix_num = len(self.prefix)
                    if prefix_num != encoding_col_num:
                        raise ValueError(
                            f"Length of 'prefix' ({prefix_num}) did not match "
                            + f"the length of the columns being encoded ({encoding_col_num})"
                        )
                    columns = self.prefix.keys()
                    for columns_columnname, prefix_columnname in zip(
                        columns, list(self.columns)
                    ):
                        if columns_columnname != prefix_columnname:
                            raise KeyError(f"{columns_columnname}")
                else:
                    self.columns = list(self.prefix.keys())
                # Convert prefix from dict to list, to simplify tile work
                self.prefix = list(self.prefix.values())

        return self.new_dataframe(
            [data],
            shape=(np.nan, np.nan),
            dtypes=None,
            index_value=data.index_value,
            columns_value=None,
        )


def get_dummies(
    data,
    prefix=None,
    prefix_sep="_",
    dummy_na=False,
    columns=None,
    sparse=False,
    drop_first=False,
    dtype=None,
):
    """
    Convert categorical variable into dummy/indicator variables.

    Parameters
    ----------
    data : array-like, Series, or DataFrame
        Data of which to get dummy indicators.
    prefix : str, list of str, or dict of str, default None
        String to append DataFrame column names.
        Pass a list with length equal to the number of columns
        when calling get_dummies on a DataFrame. Alternatively, `prefix`
        can be a dictionary mapping column names to prefixes.
    prefix_sep : str, default '_'
        If appending prefix, separator/delimiter to use. Or pass a
        list or dictionary as with `prefix`.
    dummy_na : bool, default False
        Add a column to indicate NaNs, if False NaNs are ignored.
    columns : list-like, default None
        Column names in the DataFrame to be encoded.
        If `columns` is None then all the columns with
        `object` or `category` dtype will be converted.
    sparse : bool, default False
        Whether the dummy-encoded columns should be backed by
        a :class:`SparseArray` (True) or a regular NumPy array (False).
    drop_first : bool, default False
        Whether to get k-1 dummies out of k categorical levels by removing the
        first level.
    dtype : dtype, default np.uint8
        Data type for new columns. Only a single dtype is allowed.

    Returns
    -------
    DataFrame
        Dummy-coded data.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series(list('abca'))

    >>> md.get_dummies(s).execute()
       a  b  c
    0  1  0  0
    1  0  1  0
    2  0  0  1
    3  1  0  0

    >>> s1 = ['a', 'b', np.nan]

    >>> md.get_dummies(s1).execute()
       a  b
    0  1  0
    1  0  1
    2  0  0

    >>> md.get_dummies(s1, dummy_na=True).execute()
       a  b  NaN
    0  1  0    0
    1  0  1    0
    2  0  0    1

    >>> df = md.DataFrame({'A': ['a', 'b', 'a'], 'B': ['b', 'a', 'c'],
    ...                    'C': [1, 2, 3]})

    >>> md.get_dummies(df, prefix=['col1', 'col2']).execute()
       C  col1_a  col1_b  col2_a  col2_b  col2_c
    0  1       1       0       0       1       0
    1  2       0       1       1       0       0
    2  3       1       0       0       0       1

    >>> md.get_dummies(pd.Series(list('abcaa'))).execute()
       a  b  c
    0  1  0  0
    1  0  1  0
    2  0  0  1
    3  1  0  0
    4  1  0  0

    >>> md.get_dummies(pd.Series(list('abcaa')), drop_first=True).execute()
       b  c
    0  0  0
    1  1  0
    2  0  1
    3  0  0
    4  0  0

    >>> md.get_dummies(pd.Series(list('abc')), dtype=float).execute()
         a    b    c
    0  1.0  0.0  0.0
    1  0.0  1.0  0.0
    2  0.0  0.0  1.0
    """
    if columns is not None and not isinstance(columns, list):
        raise TypeError("Input must be a list-like for parameter `columns`")

    op = DataFrameGetDummies(
        prefix, prefix_sep, dummy_na, columns, sparse, drop_first, dtype
    )

    return op(data)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Type, TypeVar, Union

import numpy as np

from .... import oscar as mo
from ....lib.aio import alru_cache
from ..core import MutableTensorInfo
from ..supervisor import MutableObjectManagerActor, MutableTensorActor
from .core import AbstractMutableAPI


APIType = TypeVar("APIType", bound="MutableAPI")


class MutableAPI(AbstractMutableAPI):
    def __init__(
        self,
        address: str,
        mutable_mananger: mo.ActorRefType[MutableObjectManagerActor],
    ):
        self._address = address
        self._mutable_manager_ref = mutable_mananger

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls, session_id: str, address: str) -> "MutableAPI":
        mutable_manager = await mo.actor_ref(
            address, MutableObjectManagerActor.gen_uid(session_id)
        )
        return MutableAPI(address, mutable_manager)

    @alru_cache(cache_exceptions=False)
    async def _get_mutable_tensor_ref(
        self, name: str
    ) -> mo.ActorRefType[MutableTensorActor]:
        return await self._mutable_manager_ref.get_mutable_tensor(name)

    async def create_mutable_tensor(
        self,
        shape: tuple,
        dtype: Union[np.dtype, str],
        name: str = None,
        default_value: Union[int, float] = 0,
        chunk_size: Union[int, Tuple] = None,
    ) -> MutableTensorInfo:
        actor_ref = await self._mutable_manager_ref.create_mutable_tensor(
            name=name,
            shape=shape,
            dtype=dtype,
            chunk_size=chunk_size,
            default_value=default_value,
        )
        return await actor_ref.info()

    @alru_cache(cache_exceptions=False)
    async def get_mutable_tensor(self, name: str):
        actor_ref = await self._mutable_manager_ref.get_mutable_tensor(name)
        return await actor_ref.info()

    async def seal_mutable_tensor(self, name: str, timestamp=None):
        # invalidate the `get_mutable_tensor` cache first.
        self.get_mutable_tensor.invalidate()
        return await self._mutable_manager_ref.seal_mutable_tensor(
            name, timestamp=timestamp
        )

    async def read(self, name: str, index, timestamp=None):
        tensor_ref = await self._get_mutable_tensor_ref(name)
        return await tensor_ref.read(index, timestamp)

    async def write(self, name: str, index, value, timestamp=None):
        tensor_ref = await self._get_mutable_tensor_ref(name)
        return await tensor_ref.write(index, value, timestamp)


class MockMutableAPI(MutableAPI):
    @classmethod
    async def create(cls: Type[APIType], session_id: str, address: str) -> "MutableAPI":
        mutable_managger = await mo.create_actor(
            MutableObjectManagerActor,
            session_id,
            address=address,
            uid=MutableObjectManagerActor.gen_uid(session_id),
        )
        return MockMutableAPI(address, mutable_managger)

    @classmethod
    async def cleanup(cls: Type[APIType], session_id: str, address: str):
        await mo.destroy_actor(
            await mo.actor_ref(address, MutableObjectManagerActor.gen_uid(session_id))
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .partition import _validate_partition_arguments, TensorPartition


def argpartition(a, kth, axis=-1, kind="introselect", order=None, **kw):
    """
    Perform an indirect partition along the given axis using the
    algorithm specified by the `kind` keyword. It returns an array of
    indices of the same shape as `a` that index data along the given
    axis in partitioned order.

    .. versionadded:: 1.8.0

    Parameters
    ----------
    a : array_like
        Tensor to sort.
    kth : int or sequence of ints
        Element index to partition by. The k-th element will be in its
        final sorted position and all smaller elements will be moved
        before it and all larger elements behind it. The order all
        elements in the partitions is undefined. If provided with a
        sequence of k-th it will partition all of them into their sorted
        position at once.
    axis : int or None, optional
        Axis along which to sort. The default is -1 (the last axis). If
        None, the flattened tensor is used.
    kind : {'introselect'}, optional
        Selection algorithm. Default is 'introselect'
    order : str or list of str, optional
        When `a` is a tensor with fields defined, this argument
        specifies which fields to compare first, second, etc. A single
        field can be specified as a string, and not all fields need be
        specified, but unspecified fields will still be used, in the
        order in which they come up in the dtype, to break ties.

    Returns
    -------
    index_tensor : Tensor, int
        Tensor of indices that partition `a` along the specified axis.
        If `a` is one-dimensional, ``a[index_tensor]`` yields a partitioned `a`.
        More generally, ``np.take_along_axis(a, index_tensor, axis=a)`` always
        yields the partitioned `a`, irrespective of dimensionality.

    See Also
    --------
    partition : Describes partition algorithms used.
    Tensor.partition : Inplace partition.
    argsort : Full indirect sort

    Notes
    -----
    See `partition` for notes on the different selection algorithms.

    Examples
    --------
    One dimensional tensor:

    >>> import mars.tensor as mt
    >>> x = mt.array([3, 4, 2, 1])
    >>> x[mt.argpartition(x, 3)].execute()
    array([2, 1, 3, 4])
    >>> x[mt.argpartition(x, (1, 3))].execute()
    array([1, 2, 3, 4])

    >>> x = [3, 4, 2, 1]
    >>> mt.array(x)[mt.argpartition(x, 3)].execute()
    array([2, 1, 3, 4])

    """
    a, kth, axis, kind, order, need_align = _validate_partition_arguments(
        a, kth, axis, kind, order, kw
    )
    op = TensorPartition(
        kth=kth,
        axis=axis,
        kind=kind,
        order=order,
        need_align=need_align,
        return_value=False,
        return_indices=True,
        dtype=a.dtype,
        gpu=a.op.gpu,
    )
    return op(a, kth)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import List

import numpy as np
import pandas as pd

try:
    import torch
    from torch.utils.data import Dataset
except ImportError:  # pragma: no cover
    torch = None
    Dataset = object

from .... import execute
from ....core.context import get_context
from ....tensor.core import TENSOR_TYPE
from ....dataframe.core import DATAFRAME_TYPE, SERIES_TYPE
from ....utils import require_not_none


ACCEPT_TYPE = (
    TENSOR_TYPE,
    DATAFRAME_TYPE,
    SERIES_TYPE,
    np.ndarray,
    pd.DataFrame,
    pd.Series,
    List,
)


@require_not_none(torch)
class MarsDataset(Dataset):
    r"""MarsDataset that inherit from torch.utils.data.Dataset.
    It converts from Mars basic datatype such as Tensor,
    DataFrame, Series. Additionally, it's constructor can receive
    np.ndarray, pd.DataFrame, pd.Series type.
    """

    def __init__(self, *tileables, fetch_kwargs=None):

        self._context = get_context()
        self._tileables = tileables
        self._fetch_kwargs = fetch_kwargs or dict()
        self._executed = False
        self._check_type()

    def _check_type(self):
        for t in self._tileables:
            if not isinstance(t, ACCEPT_TYPE):
                raise TypeError(f"Unexpected dataset type: {type(t)}")

    def _execute(self):
        execute_data = [t for t in self._tileables if isinstance(t, ACCEPT_TYPE[:3])]
        if len(execute_data):
            execute(execute_data)

    def __len__(self):
        return self._tileables[0].shape[0]

    def __getitem__(self, index):
        if not self._executed:
            self._execute()
            self._executed = True
        return tuple(self.get_data(t, index) for t in self._tileables)

    def get_data(self, t, index):
        fetch_kwargs = dict()
        if self._fetch_kwargs:
            fetch_kwargs = copy.deepcopy(self._fetch_kwargs)

        if isinstance(t, TENSOR_TYPE):
            return t[index].fetch(**fetch_kwargs)
        elif isinstance(t, np.ndarray):
            return t[index]
        elif isinstance(t, DATAFRAME_TYPE):
            return t.iloc[index].fetch(**fetch_kwargs).values
        elif isinstance(t, SERIES_TYPE):
            return t.iloc[index].fetch(**fetch_kwargs)
        elif isinstance(t, pd.DataFrame):
            return t.iloc[index].values
        elif isinstance(t, pd.Series):
            return t.iloc[index]
        else:
            return t[index]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import asyncio
import enum
import importlib
import inspect
import warnings
from typing import Dict, Iterable, List, Union

_ModulesType = Union[List, str, None]


class NodeRole(enum.Enum):
    SUPERVISOR = 0
    WORKER = 1


class AbstractService(abc.ABC):
    _instances = dict()

    def __init__(self, config: Dict, address: str):
        self._config = config
        self._address = address

    @classmethod
    def get_instance(cls, address: str, config: Dict = None):
        type_addr = (cls, address)
        if type_addr not in cls._instances:
            inst = cls._instances[type_addr] = cls(config, address)
        else:
            inst = cls._instances[type_addr]
        return inst

    @classmethod
    def clear(cls):
        cls._instances = dict()

    @abc.abstractmethod
    async def start(self):
        raise NotImplementedError

    @abc.abstractmethod
    async def stop(self):
        raise NotImplementedError

    async def create_session(self, session_id: str):
        pass

    async def destroy_session(self, session_id: str):
        pass


class EmptyService(AbstractService):
    async def start(self):
        pass

    async def stop(self):
        pass


def _find_service_entries(node_role: NodeRole, services: List, modules: List):
    svc_entries_list = []

    web_handlers = {}
    for svc_names in services:
        if isinstance(svc_names, str):
            svc_names = [svc_names]
        svc_entries = []
        for svc_name in svc_names:
            svc_mod = None
            for mod_name in modules:
                try:
                    full_mod_name = f"{mod_name}.{svc_name}.{node_role.name.lower()}"
                    svc_mod = importlib.import_module(full_mod_name)

                    abstract_derivatives = []
                    valid_derivatives = []
                    for attr_name in dir(svc_mod):
                        obj = getattr(svc_mod, attr_name)
                        if (
                            obj is not AbstractService
                            and isinstance(obj, type)
                            and issubclass(obj, AbstractService)
                        ):
                            if inspect.isabstract(obj):
                                abstract_derivatives.append(obj)
                            else:
                                valid_derivatives.append(obj)

                    svc_entries.extend(valid_derivatives)
                    if not valid_derivatives and abstract_derivatives:
                        warnings.warn(
                            f"Module {full_mod_name} does not have non-abstract "
                            f"service classes, but abstract classes "
                            f"{abstract_derivatives} found.",
                            RuntimeWarning,
                        )

                    try:
                        web_mod = importlib.import_module(
                            mod_name + "." + svc_name + ".api.web"
                        )
                        web_handlers.update(getattr(web_mod, "web_handlers", {}))
                    except ImportError:
                        pass
                except ImportError:
                    pass
            if svc_mod is None:
                raise ImportError(f"Cannot discover {node_role} for service {svc_name}")
        svc_entries_list.append(svc_entries)

    return svc_entries_list, web_handlers


def _normalize_modules(modules: _ModulesType):
    if modules is None:
        modules = []
    elif isinstance(modules, str):
        modules = [modules]
    else:
        modules = list(modules)
    modules = ["mars.services"] + modules
    return modules


def _iter_service_instances(
    node_role: NodeRole, config: Dict, address: str = None, reverse: bool = False
) -> Iterable[List[AbstractService]]:
    modules = _normalize_modules(config.get("modules"))
    service_names = config["services"]
    if reverse:
        service_names = service_names[::-1]

    svc_entries_list, _ = _find_service_entries(node_role, service_names, modules)
    for entries in svc_entries_list:
        yield [svc_entry.get_instance(address, config) for svc_entry in entries]


async def start_services(
    node_role: NodeRole, config: Dict, address: str = None, mark_ready: bool = True
):
    modules = _normalize_modules(config.get("modules"))

    # discover services
    service_names = config["services"]

    svc_entries_list, web_handlers = _find_service_entries(
        node_role, service_names, modules
    )

    if "web" in service_names:
        try:
            web_config = config["web"]
        except KeyError:
            web_config = config["web"] = dict()

        web_config["web_handlers"] = web_handlers

    for entries in svc_entries_list:
        instances = [svc_entry.get_instance(address, config) for svc_entry in entries]
        await asyncio.gather(*[inst.start() for inst in instances])

    if mark_ready and "cluster" in service_names:
        from .cluster import ClusterAPI

        cluster_api = await ClusterAPI.create(address)
        await cluster_api.mark_node_ready()


async def stop_services(node_role: NodeRole, config: Dict, address: str = None):
    for instances in _iter_service_instances(node_role, config, address, reverse=True):
        await asyncio.gather(*[inst.stop() for inst in instances])

    AbstractService.clear()


async def create_service_session(
    node_role: NodeRole, config: Dict, session_id: str = None, address: str = None
):
    for instances in _iter_service_instances(node_role, config, address):
        await asyncio.gather(*[inst.create_session(session_id) for inst in instances])


async def destroy_service_session(
    node_role: NodeRole, config: Dict, session_id: str = None, address: str = None
):
    for instances in _iter_service_instances(node_role, config, address, reverse=True):
        await asyncio.gather(*[inst.destroy_session(session_id) for inst in instances])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from sklearn.utils.extmath import row_norms as sklearn_row_norms

from ... import opcodes
from ...core import OutputType, recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import KeyField, BoolField, Int32Field
from ...tensor.array_utils import as_same_device, device, sparse
from ...tensor.core import TensorOrder
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin
from ._k_means_common import _execute_merge_update, _relocate_empty_clusters
from ._k_means_fast import update_center
from ._k_means_lloyd import update_chunk_dense, update_chunk_sparse


class KMeansLloydUpdate(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_LLOYD_UPDATE

    _x = KeyField("x")
    _sample_weight = KeyField("sample_weight")
    _x_squared_norms = KeyField("x_squared_norms")
    _centers_old = KeyField("centers_old")
    _labels = KeyField("labels")
    _update_centers = BoolField("update_centers")
    _n_clusters = Int32Field("n_clusters")

    def __init__(
        self,
        x=None,
        sample_weight=None,
        x_squared_norms=None,
        centers_old=None,
        labels=None,
        update_centers=None,
        n_clusters=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _x=x,
            _sample_weight=sample_weight,
            _x_squared_norms=x_squared_norms,
            _centers_old=centers_old,
            _labels=labels,
            _update_centers=update_centers,
            _n_clusters=n_clusters,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def x(self):
        return self._x

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def x_squared_norms(self):
        return self._x_squared_norms

    @property
    def centers_old(self):
        return self._centers_old

    @property
    def labels(self):
        return self._labels

    @property
    def update_centers(self):
        return self._update_centers

    @property
    def n_clusters(self):
        return self._n_clusters

    @property
    def output_limit(self):
        return 3 if self.stage != OperandStage.reduce else 2

    @property
    def _input_fields(self):
        return "_x", "_sample_weight", "_x_squared_norms", "_centers_old", "_labels"

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            if getattr(self, field, None) is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # labels
            self._labels.params,
            # centers_new
            {
                "shape": (self._n_clusters, self._x.shape[1]),
                "dtype": self._centers_old.dtype,
                "order": TensorOrder.C_ORDER,
            },
            # weight_in_clusters
            {
                "shape": (self._n_clusters,),
                "dtype": self._centers_old.dtype,
                "order": TensorOrder.C_ORDER,
            },
        ]
        return self.new_tileables(
            [getattr(self, field) for field in self._input_fields], kws=kws
        )

    @classmethod
    def tile(cls, op: "KMeansLloydUpdate"):
        if has_unknown_shape(*op.inputs):
            yield
        x = op.x
        if x.chunk_shape[1] != 1:  # pragma: no cover
            x = yield from recursive_tile(x.rechunk({1: x.shape[1]}))
        sample_weight = yield from recursive_tile(
            op.sample_weight.rechunk({0: x.nsplits[0]})
        )
        x_squared_norms = yield from recursive_tile(
            op.x_squared_norms.rechunk({0: x.nsplits[0]})
        )
        labels = yield from recursive_tile(op.labels.rechunk({0: x.nsplits[0]}))
        assert len(op.centers_old.chunks) == 1

        labels_chunks, centers_new_chunks, weight_in_clusters_chunks = [], [], []
        for i in range(x.chunk_shape[0]):
            x_chunk = x.cix[i, 0]
            sample_weight_chunk = sample_weight.cix[
                i,
            ]
            x_squared_norms_chunk = x_squared_norms.cix[
                i,
            ]
            labels_chunk = labels.cix[
                i,
            ]
            chunk_op = op.copy().reset_key()
            chunk_op.stage = OperandStage.map
            chunk_kws = [
                labels_chunk.params,
                {
                    "index": (0, 0),
                    "shape": (op.n_clusters, x_chunk.shape[1]),
                    "dtype": op.centers_old.dtype,
                    "order": TensorOrder.C_ORDER,
                },
                {
                    "index": (0,),
                    "shape": (op.n_clusters,),
                    "dtype": op.centers_old.dtype,
                    "order": TensorOrder.C_ORDER,
                },
            ]
            (
                labels_chunk,
                centers_new_chunk,
                weight_in_clusters_chunk,
            ) = chunk_op.new_chunks(
                [
                    x_chunk,
                    sample_weight_chunk,
                    x_squared_norms_chunk,
                    op.centers_old.chunks[0],
                    labels_chunk,
                ],
                kws=chunk_kws,
            )
            labels_chunks.append(labels_chunk)
            centers_new_chunks.append(centers_new_chunk)
            weight_in_clusters_chunks.append(weight_in_clusters_chunk)

        if op.update_centers:
            # merge centers_new and weight_in_clusters
            merge_op = KMeansLloydUpdate(
                stage=OperandStage.reduce, n_clusters=op.n_clusters
            )
            merge_chunk_kw = [
                centers_new_chunks[0].params,
                weight_in_clusters_chunks[0].params,
            ]
            centers_new_chunk, weight_in_cluster_chunk = merge_op.new_chunks(
                centers_new_chunks + weight_in_clusters_chunks, kws=merge_chunk_kw
            )
        else:
            # the data is meaningless, just pick one
            centers_new_chunk = centers_new_chunks[0]
            weight_in_cluster_chunk = weight_in_clusters_chunks[0]

        out_params = [out.params for out in op.outputs]
        # labels
        out_params[0]["nsplits"] = labels.nsplits
        out_params[0]["chunks"] = labels_chunks
        # centers_new
        out_params[1]["nsplits"] = tuple((s,) for s in op.outputs[1].shape)
        out_params[1]["chunks"] = [centers_new_chunk]
        # weight_in_clusters
        out_params[2]["nsplits"] = tuple((s,) for s in op.outputs[2].shape)
        out_params[2]["chunks"] = [weight_in_cluster_chunk]
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=out_params)

    @classmethod
    def _execute_reduce(cls, ctx, op):
        return _execute_merge_update(ctx, op)

    @classmethod
    def execute(cls, ctx, op: "KMeansLloydUpdate"):
        if op.stage == OperandStage.reduce:
            return cls._execute_reduce(ctx, op)
        else:
            (
                (x, sample_weight, x_squared_norms, centers_old, labels),
                device_id,
                xp,
            ) = as_same_device(
                [ctx[inp.key] for inp in op.inputs],
                device=op.device,
                ret_extra=True,
                copy_if_not_writeable=True,
            )

            with device(device_id):
                if not op.update_centers:
                    centers_new = centers_old.copy()
                else:
                    centers_new = np.zeros_like(centers_old)
                weight_in_clusters = np.zeros(op.n_clusters, dtype=x.dtype)
                centers_squared_norms = sklearn_row_norms(centers_old, squared=True)

                if xp is np:
                    method = update_chunk_dense
                elif xp is sparse:
                    method = update_chunk_sparse
                else:  # pragma: no cover
                    raise NotImplementedError("Does not support run on GPU")
                out_labels = labels.copy()
                method(
                    x,
                    sample_weight,
                    x_squared_norms,
                    centers_old,
                    centers_squared_norms,
                    out_labels,
                    centers_new,
                    weight_in_clusters,
                    op.update_centers,
                )

                # labels
                ctx[op.outputs[0].key] = out_labels
                # centers_new
                ctx[op.outputs[1].key] = centers_new
                # weight_in_cluster
                ctx[op.outputs[2].key] = weight_in_clusters


class KMeansLloydPostprocess(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_LLOYD_POSTPROCESS

    _centers_old = KeyField("centers_old")
    _centers_new = KeyField("centers_new")
    _center_shift = KeyField("center_shift")
    _weight_in_clusters = KeyField("weight_in_clusters")

    def __init__(
        self,
        centers_old=None,
        centers_new=None,
        center_shift=None,
        weight_in_clusters=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _centers_old=centers_old,
            _centers_new=centers_new,
            _center_shift=center_shift,
            _weight_in_clusters=weight_in_clusters,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def centers_old(self):
        return self._centers_old

    @property
    def centers_new(self):
        return self._centers_new

    @property
    def center_shift(self):
        return self._center_shift

    @property
    def weight_in_clusters(self):
        return self._weight_in_clusters

    @property
    def output_limit(self):
        return 2

    @property
    def _input_fields(self):
        return "_centers_old", "_centers_new", "_center_shift", "_weight_in_clusters"

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            ob = getattr(self, field)
            if ob is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # centers_new
            self._centers_new.params,
            # center_shift
            self._center_shift.params,
        ]
        return self.new_tileables(
            [getattr(self, f) for f in self._input_fields], kws=kws
        )

    @classmethod
    def tile(cls, op: "KMeansLloydPostprocess"):
        assert len(op.centers_old.chunks) == 1
        assert len(op.centers_new.chunks) == 1
        assert len(op.center_shift.chunks) == 1
        assert len(op.weight_in_clusters.chunks) == 1

        centers_old_chunk = op.centers_old.chunks[0]
        centers_new_chunk = op.centers_new.chunks[0]
        center_shift_chunk = op.center_shift.chunks[0]
        weight_in_clusters_chunk = op.weight_in_clusters.chunks[0]
        centers_new_chunk, center_shift_chunk = KMeansLloydPostprocess(
            centers_old=centers_old_chunk,
            centers_new=centers_new_chunk,
            center_shift=center_shift_chunk,
            weight_in_clusters=weight_in_clusters_chunk,
        ).new_chunks(
            [
                centers_old_chunk,
                centers_new_chunk,
                center_shift_chunk,
                weight_in_clusters_chunk,
            ],
            kws=[centers_new_chunk.params, center_shift_chunk.params],
        )

        centers_new_kw = op.centers_new.params
        centers_new_kw["chunks"] = [centers_new_chunk]
        centers_new_kw["nsplits"] = op.centers_new.nsplits
        center_shift_kw = op.center_shift.params
        center_shift_kw["chunks"] = [center_shift_chunk]
        center_shift_kw["nsplits"] = op.center_shift.nsplits
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[centers_new_kw, center_shift_kw])

    @classmethod
    def execute(cls, ctx, op: "KMeansLloydPostprocess"):
        (
            (centers_old, centers_new, center_shift, weight_in_clusters),
            device_id,
            xp,
        ) = as_same_device(
            [ctx[inp.key] for inp in op.inputs],
            op.device,
            ret_extra=True,
            copy_if_not_writeable=True,
        )

        with device(device_id):
            out_center_shift = center_shift.copy()
            out_centers_new = centers_new.copy()
            update_center(
                centers_old, out_centers_new, out_center_shift, weight_in_clusters
            )

            ctx[op.outputs[0].key] = out_centers_new
            ctx[op.outputs[1].key] = out_center_shift


def lloyd_iter(
    X,
    sample_weight,
    x_squared_norms,
    centers_old,
    labels,
    center_shift,
    update_centers=True,
    session=None,
    run_kwargs=None,
):
    update_op = KMeansLloydUpdate(
        x=X,
        sample_weight=sample_weight,
        x_squared_norms=x_squared_norms,
        centers_old=centers_old,
        labels=labels,
        update_centers=update_centers,
        n_clusters=centers_old.shape[0],
    )
    to_run = []
    ret = update_op()
    to_run.extend(ret)
    labels, centers_new, weight_in_clusters = ret

    if update_centers:
        centers_new, weight_in_clusters = _relocate_empty_clusters(
            X,
            sample_weight,
            centers_old,
            centers_new,
            weight_in_clusters,
            labels,
            to_run=to_run,
            session=session,
            run_kwargs=run_kwargs,
        )
        postprocess = KMeansLloydPostprocess(
            centers_old=centers_old,
            centers_new=centers_new,
            center_shift=center_shift,
            weight_in_clusters=weight_in_clusters,
        )
        centers_new, center_shift = postprocess()

    return centers_new, weight_in_clusters, labels, center_shift



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import datetime
import pickle
import uuid
from typing import List, Union

import numpy as np
import pandas as pd
import cloudpickle

from ... import opcodes as OperandDef
from ...config import options
from ...core.context import Context
from ...core.operand import OperatorLogicKeyGeneratorMixin
from ...serialization.serializables import (
    StringField,
    AnyField,
    BoolField,
    ListField,
    Int64Field,
    Float64Field,
    BytesField,
)
from ...tensor.utils import normalize_chunk_sizes
from ...typing import OperandType, TileableType
from ..arrays import ArrowStringDtype
from ..utils import parse_index, create_sa_connection, to_arrow_dtypes
from .core import (
    IncrementalIndexDatasource,
    ColumnPruneSupportedDataSourceMixin,
    IncrementalIndexDataSourceMixin,
)


class DataFrameReadSQLLogicKeyGenerator(OperatorLogicKeyGeneratorMixin):
    def _get_logic_key_token_values(self):
        fields_to_tokenize = [
            getattr(self, k, None)
            for k in [
                "table_or_sql",
                "schema",
                "coerce_float",
                "parse_dates",
                "columns",
                "method",
                "incremental_index",
                "use_arrow_dtype",
                "partition_col",
            ]
        ]
        return super()._get_logic_key_token_values() + fields_to_tokenize


class DataFrameReadSQL(
    IncrementalIndexDatasource,
    ColumnPruneSupportedDataSourceMixin,
    IncrementalIndexDataSourceMixin,
    DataFrameReadSQLLogicKeyGenerator,
):
    _op_type_ = OperandDef.READ_SQL

    table_or_sql = AnyField("table_or_sql")
    selectable = BytesField(
        "selectable", on_serialize=pickle.dumps, on_deserialize=pickle.loads
    )
    con = AnyField("con")
    schema = StringField("schema")
    index_col = AnyField("index_col")
    coerce_float = BoolField("coerce_float")
    parse_dates = AnyField("parse_dates")
    columns = ListField("columns")
    engine_kwargs = BytesField(
        "engine_kwargs",
        on_serialize=cloudpickle.dumps,
        on_deserialize=cloudpickle.loads,
    )
    row_memory_usage = Float64Field("row_memory_usage")
    method = StringField("method")
    incremental_index = BoolField("incremental_index")
    use_arrow_dtype = BoolField("use_arrow_dtype")
    chunk_size = AnyField("chunk_size")
    # for chunks
    offset = Int64Field("offset")
    partition_col = StringField("partition_col")
    num_partitions = Int64Field("num_partitions")
    low_limit = AnyField("low_limit")
    high_limit = AnyField("high_limit")
    left_end = BoolField("left_end")
    right_end = BoolField("right_end")
    nrows = Int64Field("nrows", default=None)

    def get_columns(self):
        return self.columns

    def set_pruned_columns(self, columns, *, keep_order=None):
        self.columns = columns

    def _get_selectable(self, engine_or_conn, columns=None):
        import sqlalchemy as sa
        from sqlalchemy import sql
        from sqlalchemy.exc import SQLAlchemyError

        # process table_name
        if self.selectable is not None:
            selectable = self.selectable
        else:
            if isinstance(self.table_or_sql, sa.Table):
                selectable = self.table_or_sql
                self.table_or_sql = selectable.name
            else:
                m = sa.MetaData()
                try:
                    selectable = sa.Table(
                        self.table_or_sql,
                        m,
                        autoload=True,
                        autoload_with=engine_or_conn,
                        schema=self.schema,
                    )
                except SQLAlchemyError:
                    temp_name_1 = "t1_" + binascii.b2a_hex(uuid.uuid4().bytes).decode()
                    temp_name_2 = "t2_" + binascii.b2a_hex(uuid.uuid4().bytes).decode()
                    if columns:
                        selectable = (
                            sql.text(self.table_or_sql)
                            .columns(*[sql.column(c) for c in columns])
                            .alias(temp_name_2)
                        )
                    else:
                        selectable = sql.select(
                            "*",
                            from_obj=sql.text(
                                f"({self.table_or_sql}) AS {temp_name_1}"
                            ),
                        ).alias(temp_name_2)
                    self.selectable = selectable
        return selectable

    def _collect_info(self, engine_or_conn, selectable, columns, test_rows):
        from sqlalchemy import sql

        # fetch test DataFrame
        if columns:
            query = sql.select(
                [sql.column(c) for c in columns], from_obj=selectable
            ).limit(test_rows)
        else:
            query = sql.select(selectable.columns, from_obj=selectable).limit(test_rows)
        test_df = pd.read_sql(
            query,
            engine_or_conn,
            index_col=self.index_col,
            coerce_float=self.coerce_float,
            parse_dates=self.parse_dates,
        )
        if len(test_df) == 0:
            self.row_memory_usage = None
        else:
            self.row_memory_usage = test_df.memory_usage(
                deep=True, index=True
            ).sum() / len(test_df)

        if self.method == "offset":
            # fetch size
            size = list(
                engine_or_conn.execute(
                    sql.select([sql.func.count()]).select_from(selectable)
                )
            )[0][0]
            shape = (size, test_df.shape[1])
        else:
            shape = (np.nan, test_df.shape[1])

        return test_df, shape

    def __call__(self, test_rows, chunk_size):
        import sqlalchemy as sa
        from sqlalchemy.sql import elements

        with create_sa_connection(self.con, **(self.engine_kwargs or dict())) as con:
            self.con = str(con.engine.url)
            selectable = self._get_selectable(con)

            # process index_col
            index_col = self.index_col
            if index_col is not None:
                if not isinstance(index_col, (list, tuple)):
                    index_col = (index_col,)
                new_index_col = []
                for col in index_col:
                    if isinstance(col, (sa.Column, elements.Label)):
                        new_index_col.append(col.name)
                    elif isinstance(col, str):
                        new_index_col.append(col)
                    elif col is not None:
                        raise TypeError(f"unknown index_col type: {type(col)}")
                self.index_col = new_index_col

            # process columns
            columns = self.columns or []
            new_columns = []
            for col in columns:
                if isinstance(col, str):
                    new_columns.append(col)
                else:
                    new_columns.append(col.name)
            self.columns = new_columns

            if self.columns:
                collect_cols = self.columns + (self.index_col or [])
            else:
                collect_cols = []
            test_df, shape = self._collect_info(
                con, selectable, collect_cols, test_rows
            )

            # reconstruct selectable using known column names
            if not collect_cols:
                self.columns = list(test_df.columns)
                if self.selectable is not None:
                    self.selectable = None
                    self._get_selectable(
                        con, columns=self.columns + (self.index_col or [])
                    )

            if self.method == "partition":
                if not self.index_col or self.partition_col not in self.index_col:
                    part_frame = test_df
                else:
                    part_frame = test_df.index.to_frame()

                if not issubclass(
                    part_frame[self.partition_col].dtype.type,
                    (np.number, np.datetime64),
                ):
                    raise TypeError(
                        "Type of partition column should be numeric or datetime, "
                        f"now it is {test_df[self.partition_col].dtype}"
                    )

            if isinstance(test_df.index, pd.RangeIndex):
                index_value = parse_index(
                    pd.RangeIndex(shape[0] if not np.isnan(shape[0]) else -1),
                    str(selectable),
                    self.con,
                )
            else:
                index_value = parse_index(test_df.index)

            columns_value = parse_index(test_df.columns, store_data=True)

            dtypes = test_df.dtypes
            use_arrow_dtype = self.use_arrow_dtype
            if use_arrow_dtype is None:
                use_arrow_dtype = options.dataframe.use_arrow_dtype
            if use_arrow_dtype:
                dtypes = to_arrow_dtypes(dtypes, test_df=test_df)

            return self.new_dataframe(
                None,
                shape=shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
                raw_chunk_size=chunk_size,
            )

    @classmethod
    def _tile_offset(cls, op: "DataFrameReadSQL"):
        df = op.outputs[0]

        if op.row_memory_usage is not None:
            # Data selected
            chunk_size = df.extra_params.raw_chunk_size or options.chunk_size
            if chunk_size is None:
                chunk_size = (
                    int(options.chunk_store_limit / op.row_memory_usage),
                    df.shape[1],
                )
            row_chunk_sizes = normalize_chunk_sizes(df.shape, chunk_size)[0]
        else:
            # No data selected
            row_chunk_sizes = (0,)
        offsets = np.cumsum((0,) + row_chunk_sizes).tolist()

        out_chunks = []
        for i, row_size in enumerate(row_chunk_sizes):
            chunk_op = op.copy().reset_key()
            chunk_op._row_memory_usage = None  # no need for chunk
            offset = chunk_op.offset = offsets[i]
            if df.index_value.has_value():
                # range index
                index_value = parse_index(
                    df.index_value.to_pandas()[offset : offsets[i + 1]]
                )
            else:
                index_value = parse_index(
                    df.index_value.to_pandas(),
                    op.table_or_sql or str(op.selectable),
                    op.con,
                    i,
                    row_size,
                )
            out_chunk = chunk_op.new_chunk(
                None,
                shape=(row_size, df.shape[1]),
                columns_value=df.columns_value,
                index_value=index_value,
                dtypes=df.dtypes,
                index=(i, 0),
            )
            out_chunks.append(out_chunk)

        nsplits = (row_chunk_sizes, (df.shape[1],))
        new_op = op.copy()
        return new_op.new_dataframes(
            None, chunks=out_chunks, nsplits=nsplits, **df.params
        )

    def _parse_datetime(self, val):
        if isinstance(self.parse_dates, list):
            return pd.to_datetime(val)
        args = self.parse_dates[self.partition_col]
        args = {"format": args} if isinstance(args, str) else args
        return pd.to_datetime(val, **args)

    @classmethod
    def _tile_partition(cls, op: "DataFrameReadSQL"):
        df = op.outputs[0]

        selectable = op._get_selectable(None)

        if op.low_limit is None or op.high_limit is None:
            import sqlalchemy as sa
            from sqlalchemy import sql

            engine = sa.create_engine(op.con, **(op.engine_kwargs or dict()))
            try:
                part_col = selectable.columns[op.partition_col]
                range_results = engine.execute(
                    sql.select([sql.func.min(part_col), sql.func.max(part_col)])
                )

                op.low_limit, op.high_limit = next(range_results)
                if op.parse_dates and op.partition_col in op.parse_dates:
                    op.low_limit = op._parse_datetime(op.low_limit)
                    op.high_limit = op._parse_datetime(op.high_limit)
            finally:
                engine.dispose()

        if isinstance(op.low_limit, (datetime.datetime, np.datetime64, pd.Timestamp)):
            seps = pd.date_range(op.low_limit, op.high_limit, op.num_partitions + 1)
        else:
            seps = np.linspace(
                op.low_limit, op.high_limit, op.num_partitions + 1, endpoint=True
            )

        out_chunks = []
        for i, (start, end) in enumerate(zip(seps, seps[1:])):
            chunk_op = op.copy().reset_key()
            chunk_op.row_memory_usage = None  # no need for chunk
            chunk_op.num_partitions = None
            chunk_op.low_limit = start
            chunk_op.high_limit = end
            chunk_op.left_end = i == 0
            chunk_op.right_end = i == op.num_partitions - 1

            if df.index_value.has_value():
                # range index
                index_value = parse_index(-1, chunk_op.key, chunk_op.index_value.key)
            else:
                index_value = parse_index(
                    df.index_value.to_pandas(), str(selectable), op.con, i
                )
            out_chunk = chunk_op.new_chunk(
                None,
                shape=(np.nan, df.shape[1]),
                columns_value=df.columns_value,
                index_value=index_value,
                dtypes=df.dtypes,
                index=(i, 0),
            )
            out_chunks.append(out_chunk)

        nsplits = ((np.nan,) * len(out_chunks), (df.shape[1],))
        new_op = op.copy()
        return new_op.new_dataframes(
            None, chunks=out_chunks, nsplits=nsplits, **df.params
        )

    @classmethod
    def tile(cls, op: "DataFrameReadSQL"):
        if op.method == "offset":
            return cls._tile_offset(op)
        else:
            return cls._tile_partition(op)

    @classmethod
    def post_tile(cls, op: OperandType, results: List[TileableType]):
        if op.method != "offset":
            # method `offset` knows shape of each chunk
            # just skip incremental process
            return super().post_tile(op, results)

    @classmethod
    def execute(cls, ctx, op: "DataFrameReadSQL"):
        import sqlalchemy as sa

        def _adapt_datetime(dt):
            if isinstance(dt, np.datetime64):
                return dt.astype("<M8[ms]").astype(datetime.datetime)
            elif isinstance(dt, pd.Timestamp):
                return dt.to_pydatetime()
            return dt

        out = op.outputs[0]

        engine = sa.create_engine(op.con, **(op.engine_kwargs or dict()))
        try:
            selectable = op._get_selectable(engine)

            columns = [selectable.columns[col] for col in op.columns]
            column_names = set(op.columns)
            if op.index_col:
                for icol in op.index_col:
                    if icol not in column_names:
                        columns.append(selectable.columns[icol])

            # convert to python timestamp in case np / pd time types not handled
            op.low_limit = _adapt_datetime(op.low_limit)
            op.high_limit = _adapt_datetime(op.high_limit)

            query = sa.sql.select(columns)
            if op.method == "partition":
                part_col = selectable.columns[op.partition_col]
                if op.left_end:
                    query = query.where(part_col < op.high_limit)
                elif op.right_end:
                    query = query.where(part_col >= op.low_limit)
                else:
                    query = query.where(
                        (part_col >= op.low_limit) & (part_col < op.high_limit)
                    )

            if hasattr(selectable, "primary_key") and len(selectable.primary_key) > 0:
                # if table has primary key, sort as the order
                query = query.order_by(*list(selectable.primary_key))
            elif op.index_col:
                # if no primary key, sort as the index_col
                query = query.order_by(
                    *[selectable.columns[col] for col in op.index_col]
                )
            else:
                # at last, we sort by all the columns
                query = query.order_by(*columns)

            if op.method == "offset":
                query = query.limit(out.shape[0])
                if op.offset > 0:
                    query = query.offset(op.offset)

            if op.nrows is not None:
                query = query.limit(op.nrows)

            df = pd.read_sql(
                query,
                engine,
                index_col=op.index_col,
                coerce_float=op.coerce_float,
                parse_dates=op.parse_dates,
            )
            if op.method == "offset" and op.index_col is None and op.offset > 0:
                index = pd.RangeIndex(op.offset, op.offset + out.shape[0])
                if op.nrows is not None:
                    index = index[: op.nrows]
                df.index = index

            use_arrow_dtype = op.use_arrow_dtype
            if use_arrow_dtype is None:
                use_arrow_dtype = options.dataframe.use_arrow_dtype
            if use_arrow_dtype:
                dtypes = to_arrow_dtypes(df.dtypes, test_df=df)
                for i in range(len(dtypes)):
                    dtype = dtypes.iloc[i]
                    if isinstance(dtype, ArrowStringDtype):
                        df.iloc[:, i] = df.iloc[:, i].astype(dtype)

            if out.ndim == 2:
                ctx[out.key] = df
            else:
                # this happens when column pruning results in one single series
                ctx[out.key] = df.iloc[:, 0]
        finally:
            engine.dispose()

    @classmethod
    def post_execute(cls, ctx: Union[dict, Context], op: OperandType):
        if op.method != "offset":
            # method `offset` knows shape of each chunk
            # just skip incremental process
            return super().post_execute(ctx, op)


def _read_sql(
    table_or_sql,
    con,
    schema=None,
    index_col=None,
    coerce_float=True,
    params=None,
    parse_dates=None,
    columns=None,
    chunksize=None,
    incremental_index=False,
    use_arrow_dtype=None,
    test_rows=None,
    chunk_size=None,
    engine_kwargs=None,
    partition_col=None,
    num_partitions=None,
    low_limit=None,
    high_limit=None,
):
    if chunksize is not None:
        raise NotImplementedError("read_sql_query with chunksize not supported")
    method = "offset" if partition_col is None else "partition"

    op = DataFrameReadSQL(
        table_or_sql=table_or_sql,
        selectable=None,
        con=con,
        schema=schema,
        index_col=index_col,
        coerce_float=coerce_float,
        params=params,
        parse_dates=parse_dates,
        columns=columns,
        engine_kwargs=engine_kwargs,
        incremental_index=incremental_index,
        use_arrow_dtype=use_arrow_dtype,
        method=method,
        partition_col=partition_col,
        num_partitions=num_partitions,
        low_limit=low_limit,
        high_limit=high_limit,
        chunk_size=chunk_size,
    )
    return op(test_rows, chunk_size)


def read_sql(
    sql,
    con,
    index_col=None,
    coerce_float=True,
    params=None,
    parse_dates=None,
    columns=None,
    chunksize=None,
    test_rows=5,
    chunk_size=None,
    engine_kwargs=None,
    incremental_index=True,
    partition_col=None,
    num_partitions=None,
    low_limit=None,
    high_limit=None,
):
    """
    Read SQL query or database table into a DataFrame.

    This function is a convenience wrapper around ``read_sql_table`` and
    ``read_sql_query`` (for backward compatibility). It will delegate
    to the specific function depending on the provided input. A SQL query
    will be routed to ``read_sql_query``, while a database table name will
    be routed to ``read_sql_table``. Note that the delegated function might
    have more specific notes about their functionality not listed here.

    Parameters
    ----------
    sql : str or SQLAlchemy Selectable (select or text object)
        SQL query to be executed or a table name.
    con : SQLAlchemy connectable (engine/connection) or database str URI
        or DBAPI2 connection (fallback mode)'

        Using SQLAlchemy makes it possible to use any DB supported by that
        library. If a DBAPI2 object, only sqlite3 is supported. The user is responsible
        for engine disposal and connection closure for the SQLAlchemy connectable. See
        `here <https://docs.sqlalchemy.org/en/13/core/connections.html>`_
    index_col : str or list of strings, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point, useful for SQL result sets.
    params : list, tuple or dict, optional, default: None
        List of parameters to pass to execute method.  The syntax used
        to pass parameters is database driver dependent. Check your
        database driver documentation for which of the five syntax styles,
        described in PEP 249's paramstyle, is supported.
        Eg. for psycopg2, uses %(name)s so use params={'name' : 'value'}.
    parse_dates : list or dict, default: None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times, or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    columns : list, default: None
        List of column names to select from SQL table (only used when reading
        a table).
    chunksize : int, default None
        If specified, return an iterator where `chunksize` is the number of rows
        to include in each chunk. Note that this argument is only kept for
        compatibility. If a non-none value passed, an error will be reported.
    test_rows: int, default 5
        The number of rows to fetch for inferring dtypes.
    chunk_size: : int or tuple of ints, optional
        Specifies chunk size for each dimension.
    engine_kwargs: dict, default None
        Extra kwargs to pass to sqlalchemy.create_engine
    incremental_index: bool, default True
        If index_col not specified, ensure range index incremental,
        gain a slightly better performance if setting False.
    partition_col : str, default None
        Specify name of the column to split the result of the query. If
        specified, the range ``[low_limit, high_limit]`` will be divided
        into ``n_partitions`` chunks with equal lengths. We do not
        guarantee the sizes of chunks be equal. When the value is None,
        ``OFFSET`` and ``LIMIT`` clauses will be used to cut the result
        of the query.
    num_partitions : int, default None
        The number of chunks to divide the result of the query into,
        when ``partition_col`` is specified.
    low_limit : default None
        The lower bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the minimum of
        the column.
    high_limit : default None
        The higher bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the maximum of
        the column.

    Returns
    -------
    DataFrame

    See Also
    --------
    read_sql_table : Read SQL database table into a DataFrame.
    read_sql_query : Read SQL query into a DataFrame.
    """
    return _read_sql(
        table_or_sql=sql,
        con=con,
        index_col=index_col,
        coerce_float=coerce_float,
        params=params,
        parse_dates=parse_dates,
        columns=columns,
        engine_kwargs=engine_kwargs,
        incremental_index=incremental_index,
        chunksize=chunksize,
        test_rows=test_rows,
        chunk_size=chunk_size,
        partition_col=partition_col,
        num_partitions=num_partitions,
        low_limit=low_limit,
        high_limit=high_limit,
    )


def read_sql_table(
    table_name,
    con,
    schema=None,
    index_col=None,
    coerce_float=True,
    parse_dates=None,
    columns=None,
    chunksize=None,
    test_rows=5,
    chunk_size=None,
    engine_kwargs=None,
    incremental_index=True,
    use_arrow_dtype=None,
    partition_col=None,
    num_partitions=None,
    low_limit=None,
    high_limit=None,
):
    """
    Read SQL database table into a DataFrame.

    Given a table name and a SQLAlchemy connectable, returns a DataFrame.
    This function does not support DBAPI connections.

    Parameters
    ----------
    table_name : str
        Name of SQL table in database.
    con : SQLAlchemy connectable or str
        A database URI could be provided as as str.
        SQLite DBAPI connection mode not supported.
    schema : str, default None
        Name of SQL schema in database to query (if database flavor
        supports this). Uses default schema if None (default).
    index_col : str or list of str, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point. Can result in loss of Precision.
    parse_dates : list or dict, default None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    columns : list, default None
        List of column names to select from SQL table.
    chunksize : int, default None
        If specified, returns an iterator where `chunksize` is the number of
        rows to include in each chunk. Note that this argument is only kept
        for compatibility. If a non-none value passed, an error will be
        reported.
    test_rows: int, default 5
        The number of rows to fetch for inferring dtypes.
    chunk_size: : int or tuple of ints, optional
        Specifies chunk size for each dimension.
    engine_kwargs: dict, default None
        Extra kwargs to pass to sqlalchemy.create_engine
    incremental_index: bool, default True
        If index_col not specified, ensure range index incremental,
        gain a slightly better performance if setting False.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns.
    partition_col : str, default None
        Specify name of the column to split the result of the query. If
        specified, the range ``[low_limit, high_limit]`` will be divided
        into ``n_partitions`` chunks with equal lengths. We do not
        guarantee the sizes of chunks be equal. When the value is None,
        ``OFFSET`` and ``LIMIT`` clauses will be used to cut the result
        of the query.
    num_partitions : int, default None
        The number of chunks to divide the result of the query into,
        when ``partition_col`` is specified.
    low_limit : default None
        The lower bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the minimum of
        the column.
    high_limit : default None
        The higher bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the maximum of
        the column.

    Returns
    -------
    DataFrame
        A SQL table is returned as two-dimensional data structure with labeled
        axes.

    See Also
    --------
    read_sql_query : Read SQL query into a DataFrame.
    read_sql : Read SQL query or database table into a DataFrame.

    Notes
    -----
    Any datetime values with time zone information will be converted to UTC.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> md.read_sql_table('table_name', 'postgres:///db_name')  # doctest:+SKIP
    """
    return _read_sql(
        table_or_sql=table_name,
        con=con,
        schema=schema,
        index_col=index_col,
        coerce_float=coerce_float,
        parse_dates=parse_dates,
        columns=columns,
        engine_kwargs=engine_kwargs,
        incremental_index=incremental_index,
        use_arrow_dtype=use_arrow_dtype,
        chunksize=chunksize,
        test_rows=test_rows,
        chunk_size=chunk_size,
        partition_col=partition_col,
        num_partitions=num_partitions,
        low_limit=low_limit,
        high_limit=high_limit,
    )


def read_sql_query(
    sql,
    con,
    index_col=None,
    coerce_float=True,
    params=None,
    parse_dates=None,
    chunksize=None,
    test_rows=5,
    chunk_size=None,
    engine_kwargs=None,
    incremental_index=True,
    use_arrow_dtype=None,
    partition_col=None,
    num_partitions=None,
    low_limit=None,
    high_limit=None,
):
    """
    Read SQL query into a DataFrame.

    Returns a DataFrame corresponding to the result set of the query
    string. Optionally provide an `index_col` parameter to use one of the
    columns as the index, otherwise default integer index will be used.

    Parameters
    ----------
    sql : str SQL query or SQLAlchemy Selectable (select or text object)
        SQL query to be executed.
    con : SQLAlchemy connectable(engine/connection), database str URI,
        or sqlite3 DBAPI2 connection
        Using SQLAlchemy makes it possible to use any DB supported by that
        library.
        If a DBAPI2 object, only sqlite3 is supported.
    index_col : str or list of strings, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point. Useful for SQL result sets.
    params : list, tuple or dict, optional, default: None
        List of parameters to pass to execute method.  The syntax used
        to pass parameters is database driver dependent. Check your
        database driver documentation for which of the five syntax styles,
        described in PEP 249's paramstyle, is supported.
        Eg. for psycopg2, uses %(name)s so use params={'name' : 'value'}.
    parse_dates : list or dict, default: None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times, or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    chunksize : int, default None
        If specified, return an iterator where `chunksize` is the number of
        rows to include in each chunk. Note that this argument is only kept
        for compatibility. If a non-none value passed, an error will be
        reported.
    incremental_index: bool, default True
        If index_col not specified, ensure range index incremental,
        gain a slightly better performance if setting False.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns.
    test_rows: int, default 5
        The number of rows to fetch for inferring dtypes.
    chunk_size: : int or tuple of ints, optional
        Specifies chunk size for each dimension.
    engine_kwargs: dict, default None
        Extra kwargs to pass to sqlalchemy.create_engine
    partition_col : str, default None
        Specify name of the column to split the result of the query. If
        specified, the range ``[low_limit, high_limit]`` will be divided
        into ``n_partitions`` chunks with equal lengths. We do not
        guarantee the sizes of chunks be equal. When the value is None,
        ``OFFSET`` and ``LIMIT`` clauses will be used to cut the result
        of the query.
    num_partitions : int, default None
        The number of chunks to divide the result of the query into,
        when ``partition_col`` is specified.
    low_limit : default None
        The lower bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the minimum of
        the column.
    high_limit : default None
        The higher bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the maximum of
        the column.

    Returns
    -------
    DataFrame

    See Also
    --------
    read_sql_table : Read SQL database table into a DataFrame.
    read_sql

    Notes
    -----
    Any datetime values with time zone information parsed via the `parse_dates`
    parameter will be converted to UTC.
    """
    return _read_sql(
        table_or_sql=sql,
        con=con,
        index_col=index_col,
        coerce_float=coerce_float,
        params=params,
        parse_dates=parse_dates,
        engine_kwargs=engine_kwargs,
        incremental_index=incremental_index,
        use_arrow_dtype=use_arrow_dtype,
        chunksize=chunksize,
        test_rows=test_rows,
        chunk_size=chunk_size,
        partition_col=partition_col,
        num_partitions=num_partitions,
        low_limit=low_limit,
        high_limit=high_limit,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
from collections import deque, defaultdict
from typing import Dict, List, Tuple, Type, Union

from ....config import Config
from ....core import ChunkGraph, ChunkType, enter_mode
from ....core.operand import (
    Fetch,
    VirtualOperand,
    LogicKeyGenerator,
    MapReduceOperand,
    OperandStage,
)
from ....resource import Resource
from ....typing import BandType, OperandType
from ....utils import build_fetch, tokenize
from ...subtask import SubtaskGraph, Subtask
from ..core import Task, new_task_id
from .assigner import AbstractGraphAssigner, GraphAssigner
from .fusion import Coloring

logger = logging.getLogger(__name__)


def need_reassign_worker(op: OperandType) -> bool:
    # NOTE(qinxuye): special process for reducer
    # We'd better set reducer op's stage to reduce, however,
    # in many case, we copy a reducer op from tileable op,
    # then set stage as reducer one,
    # it would be quite nasty to take over the __setattr__ and
    # make reassign_worker True etc.
    return op.reassign_worker or (
        isinstance(op, MapReduceOperand) and op.stage == OperandStage.reduce
    )


class GraphAnalyzer:
    def __init__(
        self,
        chunk_graph: ChunkGraph,
        band_resource: Dict[BandType, Resource],
        task: Task,
        config: Config,
        chunk_to_subtasks: Dict[ChunkType, Subtask],
        graph_assigner_cls: Type[AbstractGraphAssigner] = None,
        stage_id: str = None,
    ):
        self._chunk_graph = chunk_graph
        self._band_resource = band_resource
        self._task = task
        self._stage_id = stage_id
        self._config = config
        self._fuse_enabled = task.fuse_enabled
        self._extra_config = task.extra_config
        self._chunk_to_subtasks = chunk_to_subtasks
        if graph_assigner_cls is None:
            graph_assigner_cls = GraphAssigner
        self._graph_assigner_cls = graph_assigner_cls
        self._chunk_to_copied = dict()
        self._logic_key_generator = LogicKeyGenerator()

    @classmethod
    def _iter_start_ops(cls, chunk_graph: ChunkGraph):
        visited = set()
        op_keys = set()
        start_chunks = deque(chunk_graph.iter_indep())
        stack = deque([start_chunks.popleft()])

        while stack:
            chunk = stack.popleft()
            if chunk not in visited:
                inp_chunks = chunk_graph.predecessors(chunk)
                if not inp_chunks or all(
                    inp_chunk in visited for inp_chunk in inp_chunks
                ):
                    if len(inp_chunks) == 0:
                        op_key = chunk.op.key
                        if op_key not in op_keys:
                            op_keys.add(op_key)
                            yield chunk.op
                    visited.add(chunk)
                    stack.extend(c for c in chunk_graph[chunk] if c not in visited)
                else:
                    stack.appendleft(chunk)
                    stack.extendleft(
                        reversed(
                            [
                                c
                                for c in chunk_graph.predecessors(chunk)
                                if c not in visited
                            ]
                        )
                    )
            if not stack and start_chunks:
                stack.appendleft(start_chunks.popleft())

    @classmethod
    def _gen_input_chunks(
        cls,
        inp_chunks: List[ChunkType],
        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType],
    ) -> List[ChunkType]:
        # gen fetch chunks for input chunks
        inp_fetch_chunks = []
        for inp_chunk in inp_chunks:
            if inp_chunk in chunk_to_fetch_chunk:
                inp_fetch_chunks.append(chunk_to_fetch_chunk[inp_chunk])
            elif isinstance(inp_chunk.op, Fetch):
                chunk_to_fetch_chunk[inp_chunk] = inp_chunk
                inp_fetch_chunks.append(inp_chunk)
            else:
                fetch_chunk = build_fetch(inp_chunk).data
                chunk_to_fetch_chunk[inp_chunk] = fetch_chunk
                inp_fetch_chunks.append(fetch_chunk)

        return inp_fetch_chunks

    @staticmethod
    def _to_band(band_or_worker: Union[BandType, str]) -> BandType:
        if isinstance(band_or_worker, tuple) and len(band_or_worker) == 2:
            # band already
            return band_or_worker
        else:
            return band_or_worker, "numa-0"

    def _gen_subtask_info(
        self,
        chunks: List[ChunkType],
        chunk_to_subtask: Dict[ChunkType, Subtask],
        chunk_to_bands: Dict[ChunkType, BandType],
        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType],
    ) -> Tuple[Subtask, List[Subtask]]:
        # gen subtask and its input subtasks
        final_result_chunks_set = set(self._chunk_graph.result_chunks)
        chunks_set = set(chunks)

        result_chunks = []
        result_chunks_set = set()
        chunk_graph = ChunkGraph(result_chunks)
        out_of_scope_chunks = []
        chunk_to_copied = self._chunk_to_copied
        update_meta_chunks = []
        # subtask properties
        band = None
        is_virtual = None
        retryable = True
        chunk_priority = None
        expect_worker = None
        bands_specified = None
        processed = set()
        for chunk in chunks:
            if chunk in processed:
                continue
            if expect_worker is None:
                expect_worker = chunk.op.expect_worker
                bands_specified = expect_worker is not None
            else:  # pragma: no cover
                assert (
                    chunk.op.expect_worker is None
                    or expect_worker == chunk.op.expect_worker
                ), (
                    f"expect_worker {chunk.op.expect_worker} conflicts with chunks that have same color: "
                    f"{expect_worker}"
                )
            # process band
            chunk_band = chunk_to_bands.get(chunk)
            if chunk_band is not None:
                assert (
                    band is None or band == chunk_band
                ), "band conflicts with chunks that have same color"
                band = chunk_band
            # process is_virtual
            if isinstance(chunk.op, VirtualOperand):
                assert is_virtual is None, "only 1 virtual operand can exist"
                is_virtual = True
            else:
                is_virtual = False
            # process retryable
            if not chunk.op.retryable:
                retryable = False
            # process priority
            if chunk.op.priority is not None:
                assert (
                    chunk_priority is None or chunk_priority == chunk.op.priority
                ), "priority conflicts with chunks that have same color"
                chunk_priority = chunk.op.priority
            # process input chunks
            inp_chunks = []
            build_fetch_index_to_chunks = dict()
            for i, inp_chunk in enumerate(chunk.inputs):
                if inp_chunk in chunks_set:
                    inp_chunks.append(chunk_to_copied[inp_chunk])
                else:
                    build_fetch_index_to_chunks[i] = inp_chunk
                    inp_chunks.append(None)
                    if not isinstance(inp_chunk.op, Fetch):
                        out_of_scope_chunks.append(inp_chunk)
            fetch_chunks = self._gen_input_chunks(
                list(build_fetch_index_to_chunks.values()), chunk_to_fetch_chunk
            )
            for i, fetch_chunk in zip(build_fetch_index_to_chunks, fetch_chunks):
                inp_chunks[i] = fetch_chunk
            copied_op = chunk.op.copy()
            copied_op._key = chunk.op.key
            out_chunks = [
                c.data
                for c in copied_op.new_chunks(
                    inp_chunks, kws=[c.params.copy() for c in chunk.op.outputs]
                )
            ]
            for src_chunk, out_chunk in zip(chunk.op.outputs, out_chunks):
                processed.add(src_chunk)
                out_chunk._key = src_chunk.key
                chunk_graph.add_node(out_chunk)
                # cannot be copied twice
                assert src_chunk not in chunk_to_copied
                chunk_to_copied[src_chunk] = out_chunk
                if src_chunk in final_result_chunks_set:
                    if out_chunk not in result_chunks_set:
                        # add to result chunks
                        result_chunks.append(out_chunk)
                        # chunk is in the result chunks of full chunk graph
                        # meta need to be updated
                        update_meta_chunks.append(out_chunk)
                        result_chunks_set.add(out_chunk)
                if not is_virtual:
                    # skip adding fetch chunk to chunk graph when op is virtual operand
                    for c in inp_chunks:
                        if c not in chunk_graph:
                            chunk_graph.add_node(c)
                        chunk_graph.add_edge(c, out_chunk)
        # add chunks with no successors into result chunks
        result_chunks.extend(
            c
            for c in chunk_graph.iter_indep(reverse=True)
            if c not in result_chunks_set
        )
        expect_bands = (
            [self._to_band(expect_worker)]
            if bands_specified
            else ([band] if band is not None else None)
        )
        # calculate priority
        if out_of_scope_chunks:
            inp_subtasks = []
            for out_of_scope_chunk in out_of_scope_chunks:
                copied_out_of_scope_chunk = chunk_to_copied[out_of_scope_chunk]
                inp_subtask = chunk_to_subtask[out_of_scope_chunk]
                if (
                    copied_out_of_scope_chunk
                    not in inp_subtask.chunk_graph.result_chunks
                ):
                    # make sure the chunk that out of scope
                    # is in the input subtask's results,
                    # or the meta may be lost
                    inp_subtask.chunk_graph.result_chunks.append(
                        copied_out_of_scope_chunk
                    )
                inp_subtasks.append(inp_subtask)
            depth = max(st.priority[0] for st in inp_subtasks) + 1
        else:
            inp_subtasks = []
            depth = 0
        priority = (depth, chunk_priority or 0)

        subtask = Subtask(
            subtask_id=new_task_id(),
            stage_id=self._stage_id,
            logic_key=self._gen_logic_key(chunks),
            session_id=self._task.session_id,
            task_id=self._task.task_id,
            chunk_graph=chunk_graph,
            expect_bands=expect_bands,
            bands_specified=bands_specified,
            virtual=is_virtual,
            priority=priority,
            retryable=retryable,
            update_meta_chunks=update_meta_chunks,
            extra_config=self._extra_config,
        )
        return subtask, inp_subtasks

    def _gen_logic_key(self, chunks: List[ChunkType]):
        return tokenize(
            *[self._logic_key_generator.get_logic_key(chunk.op) for chunk in chunks]
        )

    @enter_mode(build=True)
    def gen_subtask_graph(
        self, op_to_bands: Dict[str, BandType] = None
    ) -> SubtaskGraph:
        """
        Analyze chunk graph and generate subtask graph.

        Returns
        -------
        subtask_graph: SubtaskGraph
            Subtask graph.
        """
        # reassign worker when specified reassign_worker = True
        # or it's a reducer operands
        reassign_worker_ops = [
            chunk.op for chunk in self._chunk_graph if need_reassign_worker(chunk.op)
        ]
        start_ops = (
            list(self._iter_start_ops(self._chunk_graph))
            if len(self._chunk_graph) > 0
            else []
        )

        # assign start chunks
        to_assign_ops = start_ops + reassign_worker_ops
        assigner = self._graph_assigner_cls(
            self._chunk_graph, to_assign_ops, self._band_resource
        )
        # assign expect workers
        cur_assigns = {
            op.key: self._to_band(op.expect_worker)
            for op in start_ops
            if op.expect_worker is not None
        }
        if op_to_bands:
            cur_assigns.update(op_to_bands)
        logger.debug(
            "Start to assign %s start chunks for task %s",
            len(start_ops),
            self._task.task_id,
        )
        chunk_to_bands = assigner.assign(cur_assigns=cur_assigns)
        logger.debug(
            "Assigned %s start chunks for task %s", len(start_ops), self._task.task_id
        )
        # assign expect workers for those specified with `expect_worker`
        # skip `start_ops`, which have been assigned before
        start_ops_set = set(start_ops)
        for chunk in self._chunk_graph:
            if chunk not in start_ops_set and chunk.op.expect_worker is not None:
                chunk_to_bands[chunk] = self._to_band(chunk.op.expect_worker)

        # color nodes
        if self._fuse_enabled:
            logger.debug("Start to fuse chunks for task %s", self._task.task_id)
            # sort start chunks in coloring as start_ops
            op_key_to_chunks = defaultdict(list)
            for chunk in self._chunk_graph:
                op_key_to_chunks[chunk.op.key].append(chunk)
            init_chunk_to_bands = dict()
            for start_op in start_ops:
                for start_chunk in op_key_to_chunks[start_op.key]:
                    init_chunk_to_bands[start_chunk] = chunk_to_bands[start_chunk]
            coloring = Coloring(
                self._chunk_graph,
                list(self._band_resource),
                init_chunk_to_bands,
                initial_same_color_num=getattr(
                    self._config, "initial_same_color_num", None
                ),
                as_broadcaster_successor_num=getattr(
                    self._config, "as_broadcaster_successor_num", None
                ),
            )
            chunk_to_colors = coloring.color()
        else:
            # if not fuse enabled, color all chunks with different colors
            op_to_colors = dict()
            chunk_to_colors = dict()
            color_gen = itertools.count()
            for c in self._chunk_graph.topological_iter():
                if c.op not in op_to_colors:
                    chunk_to_colors[c] = op_to_colors[c.op] = next(color_gen)
                else:
                    chunk_to_colors[c] = op_to_colors[c.op]
        color_to_chunks = defaultdict(list)
        for chunk, color in chunk_to_colors.items():
            if not isinstance(chunk.op, Fetch):
                color_to_chunks[color].append(chunk)

        # gen subtask graph
        subtask_graph = SubtaskGraph()
        chunk_to_fetch_chunk = dict()
        chunk_to_subtask = self._chunk_to_subtasks
        # states
        visited = set()
        logic_key_to_subtasks = defaultdict(list)
        for chunk in self._chunk_graph.topological_iter():
            if chunk in visited or isinstance(chunk.op, Fetch):
                # skip fetch chunk
                continue

            color = chunk_to_colors[chunk]
            same_color_chunks = color_to_chunks[color]
            if all(isinstance(c.op, Fetch) for c in same_color_chunks):
                # all fetch ops, no need to gen subtask
                continue
            subtask, inp_subtasks = self._gen_subtask_info(
                same_color_chunks,
                chunk_to_subtask,
                chunk_to_bands,
                chunk_to_fetch_chunk,
            )
            subtask_graph.add_node(subtask)
            logic_key_to_subtasks[subtask.logic_key].append(subtask)
            for inp_subtask in inp_subtasks:
                subtask_graph.add_edge(inp_subtask, subtask)

            for c in same_color_chunks:
                chunk_to_subtask[c] = subtask
            visited.update(same_color_chunks)

        for subtasks in logic_key_to_subtasks.values():
            for logic_index, subtask in enumerate(subtasks):
                subtask.logic_index = logic_index
                subtask.logic_parallelism = len(subtasks)
        return subtask_graph



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import AnyField
from ...tensor.rechunk.core import get_nsplits, gen_rechunk_infos, chunk_size_type
from ...typing import TileableType
from ...utils import has_unknown_shape
from ..initializer import DataFrame as asdataframe, Series as asseries, Index as asindex
from ..operands import DataFrameOperand, DataFrameOperandMixin, DATAFRAME_TYPE
from ..utils import indexing_index_value, merge_index_value


class DataFrameRechunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.RECHUNK

    chunk_size = AnyField("chunk_size")

    def __call__(self, x):
        if isinstance(x, DATAFRAME_TYPE):
            return self.new_dataframe(
                [x],
                shape=x.shape,
                dtypes=x.dtypes,
                columns_value=x.columns_value,
                index_value=x.index_value,
            )
        else:
            self.output_types = x.op.output_types
            f = (
                self.new_series
                if self.output_types[0] == OutputType.series
                else self.new_index
            )
            return f(
                [x],
                shape=x.shape,
                dtype=x.dtype,
                index_value=x.index_value,
                name=x.name,
            )

    @classmethod
    def tile(cls, op: "DataFrameRechunk"):
        from ..indexing.iloc import (
            DataFrameIlocGetItem,
            SeriesIlocGetItem,
            IndexIlocGetItem,
        )
        from ..merge.concat import DataFrameConcat

        if has_unknown_shape(*op.inputs):
            yield

        out = op.outputs[0]
        inp = op.inputs[0]
        if inp.ndim == 2:
            inp = asdataframe(inp)
        elif inp.op.output_types[0] == OutputType.series:
            inp = asseries(inp)
        else:
            inp = asindex(inp)
        chunk_size = _get_chunk_size(inp, op.chunk_size)
        if chunk_size == inp.nsplits:
            return [inp]

        rechunk_infos = gen_rechunk_infos(inp, chunk_size)
        out_chunks = []
        for rechunk_info in rechunk_infos:
            chunk_index = rechunk_info.out_index
            shape = rechunk_info.shape
            inp_chunks = rechunk_info.input_chunks
            inp_chunk_slices = rechunk_info.input_slices
            inp_slice_chunks = []
            for inp_chunk, inp_chunk_slice in zip(inp_chunks, inp_chunk_slices):
                if all(slc == slice(None) for slc in inp_chunk_slice):
                    inp_slice_chunks.append(inp_chunk)
                else:
                    index_value = indexing_index_value(
                        inp_chunk.index_value, inp_chunk_slice[0], rechunk=True
                    )
                    if inp_chunk.ndim == 1:
                        # Series or Index
                        slc_chunk_op_type = (
                            SeriesIlocGetItem
                            if op.output_types[0] == OutputType.series
                            else IndexIlocGetItem
                        )
                        slc_chunk = slc_chunk_op_type(
                            indexes=inp_chunk_slice,
                            output_types=op.output_types,
                            sparse=inp_chunk.op.sparse,
                        ).new_chunk(
                            [inp_chunk],
                            index_value=index_value,
                            dtype=inp_chunk.dtype,
                            name=inp_chunk.name,
                            index=inp_chunk.index,
                        )
                    else:
                        # DataFrame
                        columns_value = indexing_index_value(
                            inp_chunk.columns_value,
                            inp_chunk_slice[1],
                            store_data=True,
                            rechunk=True,
                        )
                        dtypes = inp_chunk.dtypes.iloc[inp_chunk_slice[1]]
                        slc_chunk = DataFrameIlocGetItem(
                            indexes=inp_chunk_slice,
                            output_types=[OutputType.dataframe],
                            sparse=inp_chunk.op.sparse,
                        ).new_chunk(
                            [inp_chunk],
                            index_value=index_value,
                            columns_value=columns_value,
                            dtypes=dtypes,
                            index=inp_chunk.index,
                        )
                    inp_slice_chunks.append(slc_chunk)

            chunk_shape = rechunk_info.input_chunk_shape
            inp_chunks_arr = np.empty(chunk_shape, dtype=object)
            inp_chunks_arr.ravel()[:] = inp_slice_chunks
            params = dict(index=chunk_index, shape=shape)
            if inp_chunks_arr.ndim == 1:
                params["index_value"] = merge_index_value(
                    {i: c.index_value for i, c in enumerate(inp_chunks_arr)}
                )
                params["dtype"] = inp_slice_chunks[0].dtype
                params["name"] = inp_slice_chunks[0].name
            else:
                params["index_value"] = merge_index_value(
                    {i: c.index_value for i, c in enumerate(inp_chunks_arr[:, 0])}
                )
                params["columns_value"] = merge_index_value(
                    {i: c.columns_value for i, c in enumerate(inp_chunks_arr[0])},
                    store_data=True,
                )
                params["dtypes"] = pd.concat([c.dtypes for c in inp_chunks_arr[0]])
            out_chunk = DataFrameConcat(
                output_types=[out.op.output_types[0]]
            ).new_chunk(inp_slice_chunks, kws=[params])
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out.params
        params["nsplits"] = chunk_size
        params["chunks"] = out_chunks
        df_or_series = new_op.new_tileable(op.inputs, kws=[params])

        if op.reassign_worker:
            for c in df_or_series.chunks:
                c.op.reassign_worker = True

        return [df_or_series]


def _get_chunk_size(
    a: TileableType, chunk_size: chunk_size_type
) -> Tuple[Tuple[int], ...]:
    if isinstance(a, DATAFRAME_TYPE):
        itemsize = max(getattr(dt, "itemsize", 8) for dt in a.dtypes)
    else:
        itemsize = a.dtype.itemsize
    return get_nsplits(a, chunk_size, itemsize)


def rechunk(a: TileableType, chunk_size: chunk_size_type, reassign_worker=False):
    if not any(pd.isna(s) for s in a.shape) and not a.is_coarse():
        if not has_unknown_shape(a):
            # do client check only when no unknown shape,
            # real nsplits will be recalculated inside `tile`
            chunk_size = _get_chunk_size(a, chunk_size)
            if chunk_size == a.nsplits:
                return a

    op = DataFrameRechunk(
        chunk_size=chunk_size,
        reassign_worker=reassign_worker,
    )
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from ....core import OperandType, TileableType, CHUNK_TYPE
from ....dataframe.base.value_counts import DataFrameValueCounts
from ....dataframe.datasource.core import HeadOptimizedDataSource
from ....dataframe.sort.core import DataFrameSortOperand
from ....dataframe.utils import parse_index
from ....utils import implements
from ..core import OptimizationRule, OptimizationRecord, OptimizationRecordType


class HeadPushDown(OptimizationRule):
    @implements(OptimizationRule.match)
    def match(self, op: OperandType) -> bool:
        node = op.outputs[0]
        input_node = self._graph.predecessors(node)[0]
        successors = self._graph.successors(input_node)
        return self._all_successor_head_pushdown(successors)

    def _all_successor_head_pushdown(self, successors: List[TileableType]):
        for succ in successors:
            rule_types = self._optimizer_cls.get_rule_types(type(succ.op))
            if rule_types is None:
                return False

            push_down_rule_types = [
                rule_type
                for rule_type in rule_types
                if issubclass(rule_type, HeadPushDown)
            ]
            if not push_down_rule_types:
                return False

            for rule_type in push_down_rule_types:
                rule = rule_type(self._graph, self._records, self._optimizer_cls)
                if not rule._can_push_down(succ.op):
                    return False
        return True

    def _can_push_down(self, op: OperandType) -> bool:
        input_nodes = self._graph.predecessors(op.outputs[0])
        accept_types = (
            HeadOptimizedDataSource,
            DataFrameSortOperand,
            DataFrameValueCounts,
        )
        if (
            len(input_nodes) == 1
            and op.can_be_optimized()
            and isinstance(input_nodes[0].op, accept_types)
            and input_nodes[0] not in self._graph.results
        ):
            return True
        return False

    def apply(self, op: OperandType):
        node = op.outputs[0]
        input_node = self._graph.predecessors(node)[0]
        nrows = input_node.op.nrows or 0
        head = op.indexes[0].stop

        new_input_op = input_node.op.copy()
        new_input_op._key = input_node.op.key
        new_input_op.nrows = nrows = max(nrows, head)
        new_input_params = input_node.params.copy()
        new_input_params["shape"] = (nrows,) + input_node.shape[1:]
        pandas_index = node.index_value.to_pandas()[:nrows]
        new_input_params["index_value"] = parse_index(pandas_index, node)
        new_input_params.update(input_node.extra_params)
        new_entity = (
            new_input_op.new_tileable
            if not isinstance(node, CHUNK_TYPE)
            else new_input_op.new_chunk
        )
        new_input_node = new_entity(input_node.inputs, kws=[new_input_params]).data

        if (
            new_input_node.op.nrows == head
            and self._graph.count_successors(input_node) == 1
        ):
            new_input_node._key = node.key
            new_input_node._id = node.id
            # just remove the input data
            self._graph.add_node(new_input_node)
            for succ in self._graph.successors(node):
                self._graph.add_edge(new_input_node, succ)
            for pred in self._graph.predecessors(input_node):
                self._graph.add_edge(pred, new_input_node)
            self._graph.remove_node(input_node)
            self._graph.remove_node(node)

            # mark optimization record
            # the input node is removed
            self._records.append_record(
                OptimizationRecord(input_node, None, OptimizationRecordType.delete)
            )
            self._records.append_record(
                OptimizationRecord(node, new_input_node, OptimizationRecordType.replace)
            )
            new_node = new_input_node
        else:
            self._replace_node(input_node, new_input_node)
            new_op = op.copy()
            new_op._key = op.key
            params = node.params.copy()
            params.update(node.extra_params)
            new_entity = (
                new_op.new_tileable
                if not isinstance(node, CHUNK_TYPE)
                else new_op.new_chunk
            )
            new_node = new_entity([new_input_node], kws=[params]).data
            self._replace_node(node, new_node)

            # mark optimization record
            self._records.append_record(
                OptimizationRecord(
                    input_node, new_input_node, OptimizationRecordType.replace
                )
            )
            self._records.append_record(
                OptimizationRecord(node, new_node, OptimizationRecordType.replace)
            )

        # check node if it's in result
        try:
            i = self._graph.results.index(node)
            self._graph.results[i] = new_node
        except ValueError:
            pass
